# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
""" Class for steel members """

import math

import numpy as np

from metku.eurocodes.en1993 import constants, en1993_1_1


class SteelMember:
    """ Steel members accroding to EN 1993
        
              
        Methods implemented in separate m files       
        % Computes the cost
        c = MemberCost(self)
        % Determine optimum profile
        %[C,sOpt] = OptimizeProfile(self,S)
        [C,sOpt] = OptimizeProfile(self,S,NLC,forceFun)


    """

    def __init__(self, profile, length, Lcr=[1.0, 1.0], mtype="beam",
                 LT_buckling=False, symmetry='dual'):
        """ Constructor
            
            profile -- member profile (of cross_section class)
            length -- member length
            lcr -- array with length 2 including the buckling
                    length factors
            type -- member type (beam, column or truss)
            
            Other attributes:
            ned -- axial force in various sections (array)
            myed -- bending moment in various sections 
                    with respect to major axis (array)
            mzed -- bending moment in various sections
                    with respect to minor axis (array)
            vyed -- shear force in various sections
                    with respect to major axis (array)
            vzed -- shear force in various sections
                    with respect to minor axis (array)
            ted -- torsion moment in various sections (array)
            loc -- locations of sections along the members in
                    local coordinates (0 <= loc[i] <= 1)
        """

        self.profile = profile
        self.length = length
        self.lcr = np.array(Lcr)
        # self.lcr[0] = Lcr[0]*length
        # self.lcr[1] = Lcr[1]*length
        self.type = mtype
        # self.ned = []
        # self.myed = []
        # self.mzed = []
        # self.vyed = []
        # self.vzed = []
        # self.ted = []
        # self.loc = []
        self.LTB = LT_buckling
        self.symmetry = symmetry

        # Dictionary of sections
        # key = local coordinate running from 0 to 1
        # value = dict with key, value pairs
        # {'ned':axial_force, 'myed':bending_moment_yaxis,
        #  'mzed': bending_moment_zaxis, 'vyed': shear force, y axis,
        #  'vzed': shear force z axis, 'ted': torsional moment,
        # 'utilization': utilization ratio in different loading cases}
        self.sections = {}

        self.__Cmy = 0.9
        self.__Cmz = 0.9
        self.__CmLT = 1.0

        self.utilization = []
        """ Utilization ratio for stability
            buck_y .. flexural buckling with respect to y axis
            buck_z .. flexural buckling with respect to z axis
            LTB .. lateral-torsional buckling
            bc_z .. beam-column stability, y axis
            bc_z .. beam-column stability, z axis
        """
        self.stability = {'buck_y': 0.0, 'buck_z': 0.0, 'LTB': 0.0, 'bc_y': 0.0, 'bc_z': 0.0}

    @property
    def ned(self):
        """ Return a list of axial forces in the sections """
        return np.array([sec['N'] for sec in self.sections.values()])

    @ned.setter
    def ned(self, val):
        """ Set the same axial force for all sections """

        for sec in self.sections.values():
            sec['N'] = val

    @property
    def myed(self):
        """ Return a list of bending moments MyEd in the sections """
        return np.array([sec['My'] for sec in self.sections.values()])

    @myed.setter
    def myed(self, val):
        for sec in self.sections.values():
            sec['My'] = val

    @property
    def mzed(self):
        """ Return a list of bending moments MzEd in the sections """
        return np.array([sec['Mz'] for sec in self.sections.values()])

    @mzed.setter
    def mzed(self, val):
        for sec in self.sections.values():
            sec['Mz'] = val

    @property
    def vyed(self):
        """ Return a list of shear forces VyEd in the sections """
        return np.array([sec['Vy'] for sec in self.sections.values()])

    @vyed.setter
    def vyed(self, val):
        for sec in self.sections.values():
            sec['Vy'] = val

    @property
    def vzed(self):
        """ Return a list of shear forces VzEd in the sections """
        return np.array([sec['Vz'] for sec in self.sections.values()])

    @vzed.setter
    def vzed(self, val):
        for sec in self.sections.values():
            sec['Vz'] = val

    @property
    def ted(self):
        """ Return a list of torques TEd in the sections """
        return np.array([sec['T'] for sec in self.sections.values()])

    @ted.setter
    def ted(self, val):
        for sec in self.sections.values():
            sec['T'] = val

    @property
    def NbRd(self):
        return np.asarray(self.buckling_strength())

    @property
    def MbRd(self):
        Mcr = self.mcrit()
        return np.asarray(self.LT_buckling_strength(Mcr))

    @property
    def MEdY(self):
        return max(abs(np.array(self.myed)))

    @property
    def MEdZ(self):
        return max(abs(np.array(self.mzed)))

    @property
    def VEdY(self):
        return max(abs(np.array(self.vyed)))
    
    @property
    def VEdZ(self):
        return max(abs(np.array(self.vzed)))

    @property
    def NEd(self):
        """ Maximum axial force (absolute value) """
        return max(abs(np.array(self.ned)))

    @property
    def NcEd(self):
        """ Maximum compressive axial force """
        return np.min(np.array(self.ned).clip(max=0.0))

    def nsect(self):
        """ Number of sections """
        # return len(self.ned)
        return len(self.sections)

    def fy(self):
        """ Yield strength of the member """
        return self.profile.fy

    def Cmy(self, sway=True, load="uniform"):
        """ Equivalent moment factor """
        if sway:
            return 0.9
        else:
            try:
                # Get end moments
                Mend = [self.sections[0]['My'], self.sections[1]['My']]
            except:
                # If no end moments are given, use the value stored in Cmy variable
                return self.__Cmy

            # Get span moment. If it is not given, assume linear distribution
            try:
                Mspan = self.sections[0.5]['My']
            except:
                Mspan = None

            self.__Cmy = en1993_1_1.equivalent_moment_factor(Mend, Mspan, load)

        return self.__Cmy

    def ncrit(self, verb=False):
        """ Buckling force according to Euler """
        C = np.pi ** 2 * self.profile.E

        ncrit = C * np.array(self.profile.I) / (self.length * self.lcr) ** 2

        if verb:
            print(f"E: {self.profile.E} MPa, Lcr: {self.length * self.lcr}, I: {self.profile.I}")
            print(f"Lcr, y: {self.lcr[0]}")
            print(f"Lcr, z: {self.lcr[1]}")
            print("Ncr,y = {0:4.2f} kN".format(ncrit[0] * 1e-3))
            print("Ncr,z = {0:4.2f} kN".format(ncrit[1] * 1e-3))

        return ncrit

    def ncrit_T(self, verb=False):
        """ Critical laod for torsional buckling """
        GIt = self.profile.G * self.profile.It
        EIw = math.pi ** 2 * self.profile.E * self.profile.Iw / self.length ** 2
        ncrit_T = 1 / self.profile.i0() ** 2 * (GIt + EIw)

        if verb:
            print("Ncr,T = {0:4.2f} kN".format(ncrit_T * 1e-3))

        return ncrit_T

    def slenderness(self, verb=False):
        """ Non-dimensional slenderness according to EN 1993-1-1 """
        NRd = self.profile.A * self.fy()
        Ncr = self.ncrit(verb)
        # if NRd <= 1e-6:
        # print(self.profile.A)
        # print(self.profile.h)
        # print(NRd, Ncr)

        slend = np.sqrt(NRd / Ncr)

        if verb:
            print("lambda,y = {0:4.3f}".format(slend[0]))
            print("lambda,z = {0:4.3f}".format(slend[1]))

        return slend

    def slenderness_torsional(self, verb=False):
        """ Non-dimensional slenderness for torsional buckling
            according to EN 1993-1-1 
        """
        NRd = self.profile.A * self.fy()
        Ncr = self.ncrit_T(verb)
        # if NRd <= 1e-6:
        # print(self.profile.A)
        # print(self.profile.h)
        # print(NRd, Ncr)

        slend = np.sqrt(NRd / Ncr)

        if verb:
            print("lambda_T = {0:4.2f}".format(slend))

        return slend

    def LT_slenderness(self, Mcr, verb=False, MRd=None):
        """ Non-dimensional slenderness for lateral-torsional buckling.
        """
        if MRd is None:
            MRd = self.profile.MRd
        lambdaLT = np.sqrt(MRd / Mcr)
        if verb:
            # print("MRd: {}, Mcr: {}".format(MRd, Mcr))
            # print("MRd0 = {0:4.2f}".format(MRd[0] * 1e-6))
            if MRd < 0:
                print(self.profile.h, self.profile.tw, self.profile.tf, self.profile.b)
            # print("MRd1 = {0:4.2f}".format(MRd[1] * 1e-6))
            # print("Mcr = {0:4.2f}".format(Mcr * 1e-6))
        return lambdaLT

    def buckling_strength(self, verb=False):
        """ Member buckling strength according to EN 1993-1-1 """

        if verb:
            print("** Buckling strength ** ")

        slend = self.slenderness(verb)
        NbRd = []
        NRd = self.profile.A * self.fy()
        # self.profile.imp_factor()
        # print(slend)
        # p = 0.5*(1+np.array(self.profile.imp_factor)*(slend-0.2)+slend**2)
        # r = 1/(p+np.sqrt(p**2-slend**2))

        for i in range(len(slend)):
            p = 0.5 * (1 + self.profile.imp_factor[i] * (slend[i] - 0.2) +
                       slend[i] ** 2)
            r = min(1 / (p + math.sqrt(p ** 2 - slend[i] ** 2)), 1.0)
            if verb:
                print("Psi,{0:1} = {1:4.3f}".format(i, p))
                print("chi,{0:1} = {1:4.3f}".format(i, r))
                print("NbRd,{0:1} = {1:4.2f} kN".format(i, NRd * r * 1e-3))

            NbRd.append(NRd * r)

        return NbRd

    def torsional_buckling_strength(self, verb=False):
        """ Member torsional buckling strength according to EN 1993-1-1 
            EN 1993-1-1, 6.3.1.4        
        """
        if verb:
            print("** Torsional buckling strength** ")

        slend = self.slenderness_torsional(verb)
        NRd = self.profile.A * self.fy()

        p = 0.5 * (1 + self.profile.imp_factor[1] * (slend - 0.2) + slend ** 2)
        r = min(1 / (p + math.sqrt(p ** 2 - slend ** 2)), 1.0)
        NbRd = r * NRd
        if verb:
            print(f"lambda_t = {slend:4.3f}")
            print(f"Psi = {p:4.2f}")
            print(f"chi = {r:4.2f}")
            print(f"NbRd = {NbRd * 1e-3:4.2f}")

        return NbRd

    def LT_buckling_strength(self, Mcr, method='specific', verb=False, MRd=None):
        """ Member lateral-torsional buckling strength """

        if MRd is None:
            MRd = self.profile.bending_resistance()[0]

        lambdaLT = self.LT_slenderness(Mcr, MRd=MRd)

        if method == 'general':
            alphaLT = self.profile.imp_factor_LT_gen
            lambdaLT0 = 0.2
            beta = 1.0
            chiLTmax = 1.0
        else:
            alphaLT = self.profile.imp_factor_LT
            lambdaLT0 = 0.4
            beta = 0.75
            chiLTmax = min(1.0, 1 / lambdaLT ** 2)

        if lambdaLT <= lambdaLT0 or self.MEdY / Mcr <= lambdaLT0 ** 2:
            chiLT = 1.0
        else:
            p = 0.5 * (1 + alphaLT * (lambdaLT - lambdaLT0) + beta * lambdaLT ** 2)
            chiLT = min(1. / (p + math.sqrt(p ** 2 - beta * lambdaLT ** 2)), chiLTmax)
        MbRd = chiLT * MRd

        if verb:
            print("lambdaLT = {0:5.4f}".format(lambdaLT))
            print("alphaLT = {0:5.4f}".format(alphaLT))
            # print("phi = {0:4.2f}".format(p))
            print("chi_LT = {0:5.4f}".format(chiLT))
            print("MbRd = {0:5.3f}".format(MbRd * 1e-6))
            print("MRd = {0:5.3f}".format(MRd * 1e-6))

        return MbRd

    def weight_per_length(self):
        """ Weight per unit length
            
            Default units: kg/mm
        """
        return self.profile.weight()

    def weight(self):
        """ Weight of the member (kg) """
        w = self.weight_per_length() * self.length
        return w

    def cost(self):
        """ Cost of member (€, $, etc.).
            By default, the cost includes only material cost.
            
            More specific cost function should be implemented for
            particular steel member using the class of that member
        """

        return self.profile.cost() * self.length

    def mcrit(self, C=[1.12, 0.45, 0.525], k=[1, 1], za=0.0, verb=False):
        """ Critical moment for lateral-torsional buckling

            For mono- and double-symmetric profiles
            C(1), C(2), C(3) -- coefficients
            za -- position of application of load
            zs -- position of the shear centre S from centroid G
            zj -- ??
            k(1) -- kz
            k(2) -- kw
        """

        kz = k[0]
        kw = k[1]

        Iz = self.profile.I[1]
        Iw = self.profile.Iw
        It = self.profile.It

        E = constants.E
        G = constants.G
        L = self.length
        C1, C2, C3 = C


        if self.symmetry == 'dual':            
            zs = 0
            zg = za - zs
            part1 = C1 * (math.pi ** 2 * E * Iz) / (kz * L) ** 2
            part2 = np.sqrt(
                (kz / kw) ** 2 * Iw / Iz +
                            (kz * L) ** 2 * G * It / (math.pi ** 2 * E * Iz) +
                            (C2 * zg) ** 2
            )
            part3 = C2 * zg

            # print("part 1 = {0:4.2f}".format(part1))
            # print("part 2 = {0:4.2f}".format(part2))
            # print("part 3 = {0:4.2f}".format(part3))

            Mcr = part1 * (part2 - part3)

        elif self.symmetry == 'mono':
            za = 0
            zs = self.profile.zs
            zg = za - zs

            zj = self.profile.zj

            part1 = C1 * (math.pi ** 2 * E * Iz) / (kz * L) ** 2
            part2 = C3 * zj - C2 * zg
            part3 = np.sqrt((kz / kw) ** 2 * Iw / Iz + (kz * L) ** 2 * G * It /
                            (math.pi ** 2 * E * Iz) + (C2 * zg - C3 * zj) ** 2)

            Mcr = part1 * (part2 + part3)

        if verb:
            print("Critical moment for LT buckling")
            print("kz = {0:4.3f}; kw = {1:4.3f}".format(kz, kw))
            print("C1 = {0:4.3f}; C2 = {1:4.3f}".format(C1, C2))
            print("Mcr = {0:6.4f} kNm".format(Mcr * 1e-6))

        # print("Iz = {0:4.2f}".format(Iz*1e-4))
        # print("Iw = {0:4.2f}".format(Iw*1e-6))
        # print("It = {0:4.2f}".format(It*1e-4))
        # print("Mcr = {0:4.2f}".format(Mcr*1e-6))
        # print("kz =", kz, "kw =", kw, "L =", L, "part3 =", part3))

        return Mcr

    def mcrit_cantilever(self, load="uniform", loc="middle", warping_free=True, verb=False):
        """ Critical buckling moment for cantilevers with I-sections 
            Based on Andrade et al. (2007), J Constr Steel Res.
            
            load .. type of load "uniform" or "point"
            loc .. location of load "middle", "top" or "bottom"
            warping .. is warping of the fixed end free
        """
        C1 = 1.0
        C2 = 0.0

        # hS is the distance between center lines of the flanges
        hS = self.profile.h - self.profile.tf

        # K is a constant that is used in the equations.
        K = np.pi / self.length * np.sqrt(
            self.profile.E * self.profile.Iz * hS ** 2 / 4 / self.profile.G / self.profile.It)

        if K < 0.1 or K > 2.5:
            print(f"Warning: constant K = {K:4.2f} is out of scope")

        denom = np.sqrt(1 + K ** 2)

        if warping_free:
            if load == 'point':
                C1 = 2.437 / denom + 0.613 * K / denom - 0.105 * K ** 2 / denom

                if loc == "top":
                    C2 = 0.409 + 1.444 * K + 0.070 * K ** 2
                elif loc == "bottom":
                    C2 = 0.529 + 0.234 * K + 0.149 * K ** 2
            else:
                C1 = 3.840 / denom + 1.496 * K / denom - 0.247 * K ** 2 / denom

                if loc == "top":
                    C2 = 0.984 + 1.420 * K + 0.165 * K ** 2
                elif loc == "bottom":
                    C2 = 1.028 + 0.388 * K + 0.150 * K ** 2
        else:
            # Warping fully restrained
            if load == 'point':
                C1 = 2.462 / denom + 2.383 * K / denom

                if loc == "top":
                    C2 = 0.380 + 2.092 * K - 0.318 * K ** 2
                elif loc == "bottom":
                    C2 = 0.512 + 0.370 * K - 0.033 * K ** 2
            else:
                C1 = 3.962 / denom + 5.531 * K / denom

                if loc == "top":
                    C2 = 1.130 + 1.539 * K - 0.176 * K ** 2
                elif loc == "bottom":
                    C2 = 1.049 + 0.234 * K - 0.020 * K ** 2

        if loc == "top":
            za = 0.5 * self.profile.h
        elif loc == "bottom":
            za = -0.5 * self.profile.h
        else:
            za = 0.0

        Mcr = self.mcrit([C1, C2, 0], [2, 1], za, verb)

        if verb:
            print(f'hS = {hS:4.2f} mm')
            print(f'Iz = {self.profile.Iz*1e-4:4.2f} cm4')
            print(f'It = {self.profile.It*1e-4:4.2f} cm4')
            print(f'Iw = {self.profile.Iw*1e-12:5.3f} dm6')
            print(f'K = {K:4.3f}')

        return Mcr

    def mcrit_buckling(self, verb=False):
        """ Evaluate elastic critical moment by buckling of compressed
            flange
        """

        # Determine compressed flange and its
        # buckling load
        _, Ifz = self.profile.compressed_flange()
        Ncrz = np.pi ** 2 * self.profile.E * Ifz / self.length ** 2

        Mcr = Ncrz * self.profile.h

        if verb:
            print("Critical moment for LT buckling using compressed flange")
            print("Ifz = {0:4.2f} mm4".format(Ifz))
            print("Ncr_z = {0:4.2f} kN".format(Ncrz * 1e-3))
            print("Mcr = {0:6.4f} kNm".format(Mcr * 1e-6))

    def add_section(self, ned=0.0, myed=0.0, mzed=0.0, vyed=0.0, vzed=0.0,
                    ted=0.0, loc=0.0):
        """ Adds new section with internal forces and location """

        self.sections[loc] = {'N': ned, 'My': myed, 'Mz': mzed,
                              'Vy': vyed, 'Vz': vzed, 'T': ted,
                              'utilization': {}}

        """
        self.ned.append(ned)
        self.myed.append(myed)
        self.mzed.append(mzed)
        self.vyed.append(vyed)
        self.vzed.append(vzed)
        self.ted.append(ted)
        self.loc.append(loc)
        self.utilization.append({})
        """

    def clear_sections(self):

        self.sections.clear()

        """
        self.ned.clear()
        self.myed.clear()
        self.mzed.clear()
        self.vyed.clear()
        self.vzed.clear()
        self.ted.clear()
        self.loc.clear()
        self.utilization.clear()
        """

    def check_section(self, loc=0, verb=False):
        """ Verify resistance of section at location 'loc' """
        self.profile.Ned = self.sections[loc]['N']
        self.profile.Med[0] = self.sections[loc]['My']
        self.profile.Ved[1] = self.sections[loc]['Vz']

        r = self.profile.section_resistance(verb=verb)

        self.sections[loc]['utilization'] = {'n': r[0], 'vz': r[1], 'my': r[2], 'mny': r[3]}
        return r

    """
    def check_section(self, n=0, verb=False):
        # Verify resistance of section 'n'
        self.profile.Ned = self.ned[n]
        self.profile.Med = self.myed[n]
        self.profile.Ved = self.vzed[n]
                

        r = self.profile.section_resistance(verb=verb)
        
        self.utilization[n] = {'n':r[0],'vz':r[1],'my':r[2],'mny':r[3]}
        return r
    """

    def check_sections(self, class1or2=True, verb=False):
        """ Verify resistance of all sections """
        r = 0.0
        for loc in self.sections.keys():
            r = max(r, max(self.check_section(loc, verb)))

        return r

        # r = np.zeros_like(self.check_section())
        # for n in range(self.nsect()):
        #    r = np.vstack((r, self.check_section(n)))
        # r.append(self.check_section(n))
        # return np.max(r, axis=0)

    def check_buckling(self):
        """ Verify buckling resistance
            output: r .. utilization ratio for buckling        
        """

        """ Find greatest compression force in the member
            clip replace all positive values with 0.0        
        """
        NEd = self.NcEd
        if NEd >= 0.0:
            """ if all axial forces are non-negative,
                buckling resistance need not be checked
            """
            r = [0.0, 0.0]
        else:
            r = abs(NEd) / self.NbRd

        self.stability['buck_y'] = r[0]
        self.stability['buck_z'] = r[1]

        return r

    def check_LT_buckling(self):
        """ Verify lateral torsional buckling resistance
            output: r .. utilization ratio for buckling        
        """
        MEd = self.MEdY
        MbRd = self.LT_buckling_strength(self.mcrit())
        r = abs(MEd) / MbRd

        return r

    def check_beamcolumn(self, sway=False, load="uniform", section_class=None, verb=False):
        """ Verify stability for combined axial force and bending moment
        """

        """ Find maximum bending moment along the member """
        MyEd = self.MEdY

        if abs(MyEd) < 1e-6:
            # If there is no bending moment, there is no need to
            # continue
            return [0, 0]

        # find the largest compression force in the member
        NEd = self.NcEd
        self.profile.Ned = NEd

        if section_class is None:
            cross_section_class = self.profile.section_class()
        else:
            cross_section_class = section_class

        # print("Check beam column:")
        # print("Section class:", cross_section_class)

        slend = self.slenderness()

        NbRd = self.buckling_strength()

        UNy = abs(NEd) / NbRd[0]
        UNz = abs(NEd) / NbRd[1]

        # Determine equivalent moment factor
        Cmy = self.Cmy(sway, load)

        kyy = en1993_1_1.kyy(UNy, slend[0], Cmy,
                             section_class=cross_section_class)

        if self.LTB:
            # Member is susceptible to lateral-torsional buckling
            kzy = en1993_1_1.kzy(kyy, UNz, slend[1], CmLT=self.CmLT,
                                 section_class=cross_section_class)
        else:
            MzEd = self.MEdZ

            if MzEd == 0:
                # Uniaxial bending: according to current EN 1993-1-1,
                # kzy = 0 is possible
                # In the revised eurocode, this is no longer valid.
                kzy = 0
            else:
                if cross_section_class < 3:
                    kzy = 0.6 * kyy
                else:
                    kzy = 0.8 * kyy

        if verb:
            print("Beam-column stability:")
            print(f'Cmy = {Cmy:4.2f}')
            print(f'kyy = {kyy:4.3f}')
            print(f'kzy = {kzy:4.3f}')

        """ For now, assume no bending with respect to z axis """
        # MzEd1 = np.max(np.array(self.mzed))
        # MzEd2 = abs(np.min(np.array(self.mzed)))
        # MzEd = max(MzEd1, MzEd2)

        # MRd = self.profile.bending_resistance(C=cross_section_class)
        MRd = self.profile.MRd

        com_comp_bend_y = -NEd / NbRd[0] + kyy * MyEd / MRd
        com_comp_bend_z = -NEd / NbRd[1] + kzy * MyEd / MRd

        com_comp_bend = [com_comp_bend_y, com_comp_bend_z]

        self.stability['bc_y'] = com_comp_bend_y
        self.stability['bc_z'] = com_comp_bend_z

        if verb:
            if com_comp_bend_y <= 1.0:
                yOK = 'OK!'
            else:
                yOK = 'NOT OK!'

            if com_comp_bend_z <= 1.0:
                zOK = 'OK!'
            else:
                zOK = 'NOT OK!'
            print(f"y-axis: {com_comp_bend_y:4.3f} <= 1.0 ({yOK})")
            print(f"z-axis: {com_comp_bend_z:4.3f} <= 1.0 ({zOK})")

        return com_comp_bend

    def required_shear_stiffness(self):
        """ EN 1993-1-1, Eq. (BB.2) """
        E = self.profile.E
        G = self.profile.G
        L = self.length
        h = self.profile.h

        Sreq = (
                       E * self.profile.Iw * math.pi ** 2 / L ** 2 + G * self.profile.It + E * self.profile.Iz * math.pi ** 2 / L ** 2 * 0.25 * h ** 2) * 70 / h ** 2

        return Sreq

    def design(self, sway=False, verb=False):
        """ Carry out member design 
            
            Returns:
                rmax .. maximum utilization ratio
        
        """

        """ Design cross sections """
        rsec = self.check_sections(class1or2=False, verb=verb)

        rmax = np.max(rsec)

        """ Check stability """
        if self.LTB:
            rLTB = self.check_LT_buckling()
            rmax = max(rmax, rLTB)

        if self.NcEd < 0.0:
            # print("Check stability")
            rbuck = self.check_buckling()
            rbeam_column = self.check_beamcolumn(sway, verb=verb)
            rmax = max(rmax, np.max(rbuck))
            rmax = max(rmax, np.max(rbeam_column))

        return rmax


"""

if __name__ == "__main__":
    
    from metku.sections.steel.ISection import HEA
    from metku.sections.steel.RHS import RHS
    
    #p = HEA(240)
    #R = RHS(200,200,8)
    p = RHS(160,160,8)
    #m = SteelMember(p,3000)
    mr = SteelMember(p,3680,Lcr=[0.9,1.0])
    mr.add_section(loc=0.0,ned=-725.57e3,myed=-22.12e6)
    mr.add_section(loc=0.5,ned=-725.57e3,myed=17.07e6)
    mr.add_section(loc=1.0,ned=-725.57e3,myed=-24.95e6)
    
    #r = mr.check_section()
    #m.ncrit(True)
    #print(m.weight())
    #m.ncrit_T(True)
    #mr.ncrit_T(True)
    #mr.ncrit(True)
"""



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 22:50:51 2018

Constants for EN 1993

@author: kmela
"""

# Constants: Young's modulus, Shear modulus, density
# Units: N, mm, i.e. MPa =N/mm^2

# Young's modulus Mpa
E = 210.0e3

# Shear modulus Mpa
G = 81.0e3

# Poisson's ratio MPa
NU = 0.3

# Density kg/mm^3
density = 7850.0e-9

# Partial safety factors
gammaM0 = 1.0
gammaM1 = 1.0
gammaM2 = 1.25
gammaM5 = 1.0

# Safety factor for fire
gammaMfi = 1.0



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
from metku.optimization.structopt import OptimizationProblem, Variable, NonLinearConstraint, LinearConstraint
from metku.frame2d.frame2d import Frame2D, SteelBeam, XYHingedSupport, XHingedSupport, LineLoad

import numpy as np

class BeamOptimizationProblem(OptimizationProblem):
    """
    Optimization example for dual-symmetric steel beam section optimization

    """
    def __init__(self):
        super().__init__(name="BeamOptimizationProblem")

        # Problem type needs to be declared, otherwise some solvers might not work
        self.prob_type = 'continuous'
        self.create_structure()
        self.create_variables()
        self.create_constraints()
        self.create_objective()


    def create_structure(self):
        """
        Creates the structure to be optimized

        In this case the structure is just the one beam
        """
        frame = Frame2D()

        # Lets create 6000 mm long beam with welded I-section profile
        coordinates = [[0, 0], [6000, 0]]
        beam = SteelBeam(coordinates,
                         profile='WI 500-12-10X300-10X300',
                         material="S355")

        # Add the beam to frame
        frame.add(beam)

        # Create supports
        sup1 = XYHingedSupport(coordinates[0])
        sup2 = XYHingedSupport(coordinates[1])

        # Add supports to frame
        frame.add(sup1)
        frame.add(sup2)

        # Create and add lineload
        ll = LineLoad(beam, [-39.15, -39.15], 'y')
        frame.add(ll)

        # Generate FE-model for the frame
        frame.generate()

        # Assing this frame as the structure to be optimized
        self.structure = frame


    def create_variables(self):
        """
        Creates optimization design variables

        In this case the variables are cross-section's dimensions
        """
        # Initialize as an empty array
        self.vars = []

        # Get the beam
        beam = self.structure.members[0]

        # Get the SteelSection object from beam
        section = beam.cross_section

        # CREATING DESIGN VARIABLES
        # -------------------------
        # Flanges' breadth
        var_BF = Variable(name="BF",
                     lb=100,
                     ub=700,
                     target={"property": "BF",
                             'objects': [section]})
        
        # Flanges' thickness
        var_TF = Variable(name="TF",
                     lb=5,
                     ub=100,
                     target={"property": "TF",
                             'objects': [section]})
        # Section's height
        var_H = Variable(name="H",
                     lb=300,
                     ub=3300,
                     target={"property": "H",
                             'objects': [section]})
        # Web's thickness
        var_TW = Variable(name="TW",
                     lb=6,
                     ub=100,
                     target={"property": "TW",
                             'objects': [section]})

        # Append created variables to list
        # This is the order of the design variables
        self.vars = [var_BF, var_TF, var_H, var_TW]




    def create_confuns(self, beam):
        """
        Creates constraint functions

        :param beam: FrameMember object
        :return: constraint functions
        """
        def bending_moment_confun(x):
            # Index 0 is the bending resistance about the major axis
            # g(x) <= 0 
            # MEd <= MRd(x) 
            g = beam.med / beam.MRd[0] - 1

            return g

        def shear_confun(x):
            g = beam.ved / beam.VRd - 1

            return g

        def deflection_confun(x):

            min_v = min([disp[1] for disp in beam.nodal_displacements.values()])
            max_v = max([disp[1] for disp in beam.nodal_displacements.values()])
            abs_max_v = max(abs(min_v), max_v)

            g = abs_max_v / (beam.length / 250) - 1

            return g

        return bending_moment_confun, shear_confun, deflection_confun

    def flange_class_con(self):
        """
        Creates flange class constraint
        :return: constraint
        """
        # Get the beam
        beam = self.structure.members[0]
        eps = np.sqrt(235 / beam.fy)

        # The order of the variables is defined in create variables -method
        # x = BF, TF, H, TW

        # NOTE: This neglects the weld
        a = np.array([0.5, -14*eps, 0, -0.5])

        con = LinearConstraint(a=a,
                               b=0,
                               con_type='<',
                               name="Flange class")
        return con


    def web_class_con(self):
        """
        Creates flange class constraint
        :return: constraint
        """
        # Get the beam
        beam = self.structure.members[0]
        eps = np.sqrt(235 / beam.fy)

        # The order of the variables is defined in create variables -method
        # BF, TF, H, TW = x

        # NOTE: This neglects the weld
        a = np.array([0, -1, 1, -124*eps])

        con = LinearConstraint(a=a,
                               b=0,
                               con_type='<',
                               name="Web class")
        return con

    def create_constraints(self):
        """
        Creates constraints

        """
        # Initialize as an empty array
        self.cons = []

        # Get the beam
        beam = self.structure.members[0]

        # Create constraint functions
        bending_moment_confun, shear_confun, deflection_confun = self.create_confuns(beam)

        # Create constraints
        bending_constr = NonLinearConstraint(con_fun=bending_moment_confun,
                                             name="Bending Constraint",
                                             parent=self)

        shear_constr = NonLinearConstraint(con_fun=shear_confun,
                                             name="Shear Constraint",
                                             parent=self)

        deflection_constr = NonLinearConstraint(con_fun=deflection_confun,
                                             name="Deflection Constraint",
                                             parent=self)
        # Deflection requires FEM analysis
        deflection_constr.fea_required = True

        # Cross-section class constraints
        web_con = self.web_class_con()
        flange_con = self.flange_class_con()

        # Assign constraints
        self.cons = [bending_constr,
                     shear_constr,
                     deflection_constr,
                     web_con,
                     flange_con]


    def create_objective(self):
        """
        Creates the objective function, function to be minimized

        In this case the objective is to minimize beam's weight
        """
        def obj(x):
            BF, TF, H, TW = x

            beam = self.structure.members[0]
            return beam.weight

        self.obj = obj


if __name__ == '__main__':

    from metku.optimization.solvers import *

    problem = BeamOptimizationProblem()

    x0 = [var.lb for var in problem.vars]
    solver = SLSQP()

    xopt, fopt = solver.solve(problem, x0=x0, maxiter=20)

    problem(xopt)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
@author Viktor Haimi
"""
import math
import sys
try:
    from metku.eurocodes.en1995.fire_protection import FireProtection
except:
    from eurocodes.en1995.fire_protection import FireProtection

if sys.version_info[0] == 3 and sys.version_info[1] < 9:
    from typing import List
    list_out = List[float]
else:
    list_out = list[float]


def charring_speed(rhok: (int, float), type: str, hardness: str, h_p: (int, float)) -> list_out:
    """
    Ensin tarkistaa RIL-205-2-2009 Taulukon 3.2 mukaisesti ja
    jos ei löydy tai paksuus on alle 20mm lasketaan beta_0_rho_t Kaavan 3.4 mukaan
    käyttää timber_datan rhok:ta eikä rhomeaniä

    @param rhok: ominaistiheys
    @param type: vaihtoehdot: solid_timber, glt, lvl
    @param hardness: vaihtoehdot: hardwood, softwood
    @param h_p: paksuus
    @return: list [beta_0, beta_n]
    """
    beta = None

    if type == 'solid_timber' or type == 'glt':
        if hardness == 'softwood':
            if type == 'solid_timber':
                if rhok >= 290:
                    beta = [0.65, 0.8]
            elif type == 'glt':
                if rhok >= 290:
                    beta = [0.65, 0.7]
        elif hardness == 'hardwood':
            if rhok == 290:
                beta = [0.65, 0.7]
            if rhok >= 450:
                beta = [0.5, 0.55]
    elif type == 'lvl':
        if rhok >= 480:
            beta = [0.65, 0.7]
        elif rhok >= 410:
            beta = [0.7, 0.75]
    else:
        beta = [1.0, 1.0]

    if h_p < 20:
        krho = math.sqrt(450 / rhok)
        k_h = math.sqrt(20 / h_p)
        beta = [krho * k_h * beta[0], beta[1]]

    return beta


def d_char_n(fp: FireProtection, t: (int, float), beta_n: float) -> float:
    """
    Returns the burned depth of single side
    @param fp: fire protection on the side
    @param t: how long the fire has burned
    @param beta_n: charring speed
    @return: depth of charring
    """

    k0 = 1.0
    d0 = 7
    # RIL-205-2-2009 Taulukko 4.1
    if fp:
        if fp.t_ch <= 20:
            if t < 20:
                k0 = t / 20
        else:
            if t <= fp.t_ch:
                k0 = t/fp.t_ch
    else:
        if t < 20:
            k0 = t / 20

    if not fp:
        return t * beta_n + k0 * d0
    # RIL-205-2-2009 kuvat 3.3 - 3.5
    if t < fp.t_ch:
        return 0
    if fp.t_ch < fp.t_f:
        if t < fp.t_f:
            return (t - fp.t_ch) * 0.73 * beta_n + k0 * d0
        elif fp.t_f < t < fp.t_a:
            return (fp.t_f - fp.t_ch) * 0.73 * beta_n + (t - fp.t_f) * 2 * beta_n + k0 * d0
        else:
            return (fp.t_f - fp.t_ch) * 0.73 * beta_n + (fp.t_a - fp.t_f) * 2 * beta_n + (t - fp.t_a) * beta_n + k0 * d0
    if t < fp.t_a:
        dt = t - fp.t_f
        return 2 * beta_n * dt + k0 * d0
    else:
        t_fa = fp.t_a - fp.t_f
        dt = t - fp.t_a
        return 2 * beta_n * t_fa + beta_n * dt + k0 * d0


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Mon Jan  4 12:45:41 2021

Constants from EN 1995

@author: Viktor Haimi
"""
# Partial safety factors taken from EN 1995-1-1 § 2, Table 2.3
gammaM_st = 1.3
gammaM_glt = 1.25
gammaM_lvl = 1.2
gammaM_other = 1.3

# SFS-EN 1995-1-1 § 6.1.6
k_m = 0.7



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 20 14:39:18 2022

Method of Moving Asymptotes

@author: kmela
"""

from copy import copy

import numpy as np

import metku.optimization.constraints as cons
from metku.optimization.opt_functions import mma_fun, NumGrad
from metku.optimization.solvers.optsolver import OptSolver
from metku.optimization.solvers.slsqp import SLSQP
from metku.optimization.structopt import OptimizationProblem
from metku.optimization.variables import Variable


class MMAProblem(OptimizationProblem):
    """ Class for MMA subproblems """

    def __init__(self, P: OptimizationProblem,
                 x: np.array,
                 U: np.array,
                 L: np.array):
        """
        Generates MMA approximation of the problem at point 'x' using the asymptotes 'U' and 'L'


        Parameters
        ----------
        x : numpy array
            Approximation point.
        U : numpy array
            Upper asymptote.
        L : numpy array
            Lower asymptote.

        Returns
        -------
        Optimization problem with nonlinear function replaced by their MMA approximations.

        """

        super().__init__(name="MMA", structure=P.structure)

        self.xk = x
        self.U = [U]
        self.L = [L]

        for i, var in enumerate(P.vars):
            # Create variables for the mma problem
            # Rules for variable bounds are from Svanberg (2007)
            lb = max(var.lb, L[i] + 0.1 * (x[i] - abs(L[i])), x[i] - 0.5 * (var.ub - var.lb))
            ub = min(var.ub, U[i] - 0.1 * (U[i] - x[i]), x[i] + 0.5 * (var.ub - var.lb))
            self.add(Variable(name=var.name,
                              lb=lb,
                              ub=ub,
                              value=x[i]))

        # Approximate objective
        if P.grad is None:
            dfx = NumGrad(P.obj, x)
        else:
            dfx = P.grad(x)

        self.obj = mma_fun(x, P.obj(x), dfx, L=L, U=U)

        # Approximate constraints
        self.cons.clear()
        for con in P.cons:
            if isinstance(con, cons.LinearConstraint):
                # Linear constraints are simply added
                self.add(con)
            else:
                # For nonlinear constraints, MMA approximation is made                
                mma_con = copy(con)
                gx = con(x)
                dgx = con.df(x)
                mma_con.con = mma_fun(x, gx, dgx, L, U)
                mma_con.fea_required = False

                self.add(mma_con)


class MMA(OptSolver):

    def __init__(self):
        super().__init__()

        self.L = []
        self.U = []
        self.mma_probs = []

    def gamma(self, xk, xk1, xk2):
        """ Parameter for modifying the asymptotes 
            xk .. current iterate
            xk1 .. x(k-1)
            xk2 .. x(k-2)
        """
        xtol = 1e-6
        xprod = (xk - xk1) * (xk1 - xk2)
        return np.array([0.7 if x < -xtol else 1.2 if x > xtol else 1 for x in xprod])

    def take_action(self):
        """ Make the MMA sub-problem and solve it. To create the sub-problem
            means the following:
                
                1. Make MMA approximations for each function at the current
                   iterate
                2. Define asymptotes L and U for each variable. Here,
                   variable bounds, asymptotes from previous iterations
                   and previous iterates are needed.
        """
        # print("MMA ACTION")
        # print(self.xvals)
        iteration = len(self.xvals) - 1
        xk = self.xvals[-1]
        xlb = self.problem.xlb()
        xub = self.problem.xub()
        Dx = xub - xlb

        # print(iteration,xk,Dx)

        # Update asymptotes (Svanberg (2007))
        if iteration <= 1:
            Lnew = xk - 0.5 * Dx
            Unew = xk + 0.5 * Dx
            # print("L = ",Lnew)
            # print("U = ",Unew)
        else:
            xk1 = self.xvals[-2]
            g = self.gamma(xk, xk1, self.xvals[-3])
            Lnew = xk - g * (xk1 - self.L[-1])
            Unew = xk + g * (self.U[-1] - xk1)

        Lnew = np.minimum(np.maximum(xk - 10 * Dx, Lnew), xk - 0.1 * Dx)
        Unew = np.minimum(np.maximum(xk + 0.01 * Dx, Unew), xk + 10 * Dx)

        # print("Lnew = ",Lnew)
        # print("Unew = ",Unew)

        self.L.append(Lnew)
        self.U.append(Unew)

        # Create MMA sub-problem
        Pmma = MMAProblem(self.problem, x=xk, U=Unew, L=Lnew)
        self.mma_probs.append(Pmma)

        # Solve the MMA sub-problem with SQP.
        # Here, a separate solver for the MMA problem could be implemented
        sqp_solver = SLSQP()
        fmma, xmma = sqp_solver.solve(Pmma, x0=xk, verb=False, maxiter=50)

        # Pmma(xmma)
        # print(Pmma.cons)
        # print(Pmma.xlb(),Pmma.xub())

        return xmma

    def step(self, action):
        """ Update design varible values """
        self.X = action

        self.problem.substitute_variables(self.X)

        return self.X.copy(), 1, False, 'INFO'



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Mon Mar 26 16:06:13 2018

Cold-formed profiles

@author: kmela
"""

from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
from metku.materials.steel_data import Steel
from metku.eurocodes.en1993.en1993_1_3.open_prof import OpenProf
from metku.eurocodes.en1993.en1993_1_3 import en1993_1_3
from metku.eurocodes.en1993 import en1993_1_5
from metku.eurocodes.en1993.constants import gammaM0, gammaM1

weckman_c_z = {100:{'a':42,'b':48,'c':18,'r':4,'t':[1,1.2,1.5,2],'width':[216,216,212,210]},
             120:{'a':42,'b':48,'c':18,'r':4,'t':[1,1.2,1.5,2],'width':[238,236,234,230]},
             150:{'a':42,'b':48,'c':18,'r':4,'t':[1.2,1.5,2],'width':[266,264,260]},
             200:{'a':42,'b':48,'c':18,'r':4,'t':[1.5,2,2.5],'width':[312,310,305]},
             250:{'a':65,'b':73,'c':21.5,'r':4,'t':[1.5,2,2.5],'width':[419,415,411]}
             }

def ruukki_z(h,t):
    
    if h <= 150:
        c = 18.0
        r = 2.0
        if t == 1.0:
            a = 45.0
            b = 39.0
        elif t == 1.2:
            a = 45.4
            b = 39.4
        elif t == 1.5:
            a = 46.0
            b = 40.0
        elif t == 2.0:
            a = 47.0
            b = 41.0
    elif h <= 250:
        c = 26.0
        r = 2.0
        if t == 1.5:
            a = 70.0
            b = 62.0
        elif t == 2.0:
            a = 71.0
            b = 63.0
        elif t == 2.5:
            a = 72.0
            b = 64.0
        elif t == 3.0:
            a = 73.0
            b = 65.0
    elif h == 300:
        c = 26.0
        r = 1.0
        if t == 1.5:
            a = 89.0
            b = 81.0
        elif t == 2.0:
            a = 90.0
            b = 82.0
        elif t == 2.5:
            a = 91.0
            b = 83.0
        elif t == 3.0:
            a = 92.0
            b = 84.0
    elif h == 350:
        c = 30.0
        r = 2.0
        if t == 2.0:
            a = 90.0
            b = 82.0
        elif t == 2.5:
            a = 91.0
            b = 83.0
        elif t == 3.0:
            a = 92
            b = 84
            
    z = ZSection(t_nom=t,h=h,a=a,b=b,ca=c,cb=c,r=r)
    
    return z
        
def corner_points(xc,r,phi=[0,np.pi],n=6):
    """ Creates corner points
        :param xc: center point
        :param r: radius of the circle
        :param phi: [start_angle,end_angle] (radians)
        :param n: number of division points
    """
    angles = np.linspace(phi[0],phi[1],n)
    return [xc + r*np.array([np.cos(angle),np.sin(angle)]) for angle in angles]


class ZSection:
    """ Class for unstiffened Z-sections """
    
    def __init__(self,t_nom=2.0,h=120,a=60,b=60,ca=20,cb=20,r=2.0,
                 material="S350GD",t_coat=0.04,
                 MEd=0.0, NEd=0.0, VEd=0.0,corners=True):
        """ Constructor 
            input:
                t_nom .. nominal wall thickness
                h .. total height
                a .. width of top flange
                b .. width of bottom flange                
                ca .. length of top flange edge stiffener (=0 if no stiffener)
                cb .. length of bottom flange edge stiffener (=0 if no stiffener)
                r .. rounding
                material .. 
                t_coat = 0.04 .. coating [mm]        
        """
        
        """ Flag: if True, then rounded corners are taken in to account in cross-section
            modelling. if False, then cross-section is modelled using centerline intersection
            points.
        """
        self.rounded_corners = corners
        self.t_nom = t_nom
        self._h = h
        self._a = a
        self._b = b
        self._ca = ca
        self._cb = cb
        self.material = Steel(material)
        self.t_coat = t_coat
        self.t_core = t_nom-t_coat
        
        self.Ned = NEd
        self.Med = MEd
        self.Ved = VEd
        
        self._points = []
        
        """ Reduction factors for effective cross-section """
        self._lambda_t = 0.1
        self._rho_t = 1.0
        self._lambda_t_lip = 0.1
        self._rho_t_lip = 1.0
        self._lambda_w = 0.1
        self._rho_w = 1.0
        self._psi_w = 1.0 # Stress ratio for web
        self._lambda_b = 0.1
        self._rho_b = 1.0
        self._lambda_b_lip = 0.1
        self._rho_b_lip = 1.0
        """ Reduction factors for distortional buckling """
        self._chi_d_t = 1.0
        self._chi_d_b = 1.0
        
        if r == 0:
            """ According to EN 10162, Table A.2;
                S320GD+Z and S350GD+Z
            """
            self.r = 1.5*t_nom
        else:
            self.r = r
            
        
        if corners is False: 
            if self.r/self.t <= 5 and self.r/self.bp_t <= 0.1 and self.r/self.bp_b <= 0.1:
                self.rounded_corners = False
            else:
                self.rounded_corners = True
        
        """ Calculation models for gross cross section and
            effective cross section.
        """
        
        self.parts = {'top_flange':{'nodes':[],'segments':[]},
                      'web':{'nodes':[],'segments':[]},
                      'bottom_flange':{'nodes':[],'segments':[]},                      
                      'top_lip':{'nodes':[],'segments':[]},
                      'bottom_lip':{'nodes':[],'segments':[]},
                      }
        
        self.prof = None
        self.create_points(n=5)
        self.create_model()
        self.prof_eff = None
        
        """ Calculation models for effective top and bottom
            edge stiffeners
        """
        self.edge_stiff = {'top':None, 'bottom':None}
        
        self.free_flange = {'top':None, 'bottom':None}
    
    
    def issymmetric(self):
        """ Cross-section is considered symmetric if flanges have
            equal widths and edge stiffeners have equal lengths
        """
        return self.a == self.b and self.ca == self.cb
    
    @property
    def npoints(self):
        """ Number of points in the section """
        
        return len(self._points)
    
    @property
    def Ro(self):
        """ Outer radius """
        return self.r + self.t_nom
    
    @property
    def Rm(self):
        """ Mid radius """
        return self.r + 0.5*self.t_nom
    
    @property
    def E(self):
        """ Young's modulus """
        return self.material.E
    
    @property
    def G(self):
        """ Shear modulus """
        return self.material.G
    
    @property
    def A(self):
        """ Gross area of cross section """
        return self.area()

    @property
    def Aeff(self):
        self.prof_eff = deepcopy(self.prof)
        self.effective_section('compression')
        return self.prof_eff.area()
    
    @property
    def Iw(self):
        """ Warping constant of the gross cross-section """
        return self.prof.warping_constant()
    
    @property
    def It(self):
        """ Torsion constant """
        return self.prof.torsion_constant()
    
    @property
    def Iy(self):
        """ Second moment of area of gross cross-section with respect
            to the major axis
        """
        return self.prof.Iy()
    
    @property
    def Iz(self):
        """ Second moment of area of gross cross-section with respect
            to the minor axis
        """
        return self.prof.Iz()
    
    @property
    def Iyz(self):
        """ Product moment of the gross cross-section """
        return self.prof.product_moment_gc()
    
    def Ifz(self,flange='bottom'):
        """ Second moment of area of the free flange """
        return self.free_flange[flange].Iz()
    
    @property
    def NRd(self):
        return self.axial_force_resistance()
    
    @property
    def fya(self):
        """ Average yield strength """        
        return en1993_1_3.fya(self.fyb,self.fu,self.t,Ag=self.Agross(),n=4)
        
    @property
    def fyb(self):
        """ Basic yield strength """
        return self.material.fy
    
    @property
    def fu(self):
        """ Ultimate strengt """
        return self.material.fu
    
    @property
    def t(self):
        """ Design thickness of the profile """
        return self.t_core
    
    @property
    def h(self):
        """ Design value of the height of the profile """
        return self._h
        #return self.design_value('_h')
    
    @property
    def a(self):
        """ Design value of the width of the top flange """
        return self._a
        #return self.design_value('_a')
    
    @property
    def b(self):
        """ Design value of the width of the bottom flange """
        return self._b
        #return self.design_value('_b')
    
    @property
    def ca(self):
        """ Design value of the top flange edge stiffener """
        return self._ca
        #return self.design_value('_ca')
    
    @property
    def cb(self):
        """ Design value of the bottom flange edge stiffener """
        return self._cb
        #return self.design_value('_cb')
    
    @property
    def bp_w(self):
        """ Notional width of the web """
        if self.rounded_corners:
            return en1993_1_3.notional_width(self.hw,[self.rm,self.rm],[90,90])
        else:
            return self.hw_center
    
    @property
    def bp_t(self):
        """ Notional width of the top flange """
        if self.rounded_corners:
            if self.ca > 0.0:            
                    bp = en1993_1_3.notional_width(self.c_top,[self.rm,self.rm],[90,90])
            else:
                    bp = en1993_1_3.notional_width(self.c_top,[self.rm],[90])            
        else:
            bp = self.bt_center
        return bp
    
    @property
    def bp_b(self):
        """ Notional width of the bottom flange """
        if self.cb > 0.0:
            if self.rounded_corners:
                bp = en1993_1_3.notional_width(self.c_bottom,[self.rm,self.rm],[90,90])
            else:
                bp = self.bb_center
        else:
            if self.rounded_corners:
                bp = en1993_1_3.notional_width(self.c_bottom,[self.rm],[90])
            else:
                bp = self.bb_center
        return bp
    
    @property
    def bp_t_lip(self):
        """ Notional width of the top flange lip"""
        if self.rounded_corners:
            if self.ca > 0.0:
                bp = en1993_1_3.notional_width(self.ca_straight,[self.rm],[90])
            else:
                bp = 0.0
        else:
            bp = self.bt_lip_center
        return bp
    
    @property
    def bp_b_lip(self):
        """ Notional width of the bottom flange lip """
        if self.rounded_corners:
            if self.cb > 0.0:
                bp = en1993_1_3.notional_width(self.cb_straight,[self.rm],[90])
            else:
                bp = 0.0
        else:
            bp = self.bb_lip_center
        return bp
    
    @property
    def shear_center(self):
        """ Coordinates of the shear center """
        y_sc, z_sc = self.prof.shear_center()
        return [y_sc, z_sc]
    
    
    
    @property
    def gr(self):
        """ Dimension gr in the rounded corner """
        return en1993_1_3.gr(self.r,self.t_nom,45)
    
    def add(self,this,part=None):
        """ Adds stuff to the section """
        
        if isinstance(this,list) or isinstance(this,np.ndarray):
            """ Item to be added is a coordinate point """
            self._points.append(this)
            if part is not None:
                self.parts[part]["nodes"].append(self.npoints-1)
    
    def design_value(self,attr):
        """ Calculates the design value of a dimension 
            This means that the coating is reduced from the dimension.
        """
        nom_val = getattr(self,attr)        
        return nom_val-self.t_coat
    
    def weight(self):
        """ Weight per unit length (kg/mm) """
        
        return self.material.rho * self.Agross()
    
    def self_weight(self):
        """ Self-weight (kN/m) """
        
        return 9.81*self.weight()
    
    def create_points(self,n=5):
        """ Creates points for the calculation model 
            :param n: number of divisions in the corner areas
            
            Number starts from the end of the flange of the extension
            and goes counter-clockwise
        """
        # Number of points; used as an index
        npt = 0
        
        self.corner_pts = n
        h = self.h
        hweb = self.hw
        
        tnom = self.t_nom
        rm = self.Rm
        ro = self.Ro
        r = self.r
        
        bp_web = self.hw_center
        bp_bottom = self.bb_center
        bp_top = self.bt_center
        bp_bottom_lip = self.bb_lip_center
        bp_top_lip = self.bt_lip_center
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
       
        if bp_bottom_lip > 0.0:
            bottom_lip_nodes = [[-bp_bottom,web_nodes[0][1]+bp_bottom_lip],
                                [-bp_bottom,web_nodes[0][1]+bp_bottom_lip-0.5*self.cb_straight]]
            for node in bottom_lip_nodes:
                self.add(node,part='bottom_lip')
            
            # Corner points
            # Lip-to-flange corner, bottom
            x_bottom_lip = (-bp_bottom+rm,web_nodes[0][1]+rm)
        
            newPts = corner_points(x_bottom_lip,rm,phi=[np.pi,1.5*np.pi],n=n)
        
            for newPt in newPts[:-1]:            
                self.add(newPt)
            
            # The second and third nodes are for helping to cope
            # with the effective cross section.
            bottom_nodes = [[-bp_bottom+rm,web_nodes[0][1]],
                            [-bp_bottom+rm+1/3*self.c_bottom,web_nodes[0][1]],
                            [-bp_bottom+rm+2/3*self.c_bottom,web_nodes[0][1]],
                            [-bp_bottom+rm+self.c_bottom,web_nodes[0][1]]
                            ]            
        else:
            bottom_nodes = [[-bp_bottom,web_nodes[0][1]],
                            [-bp_bottom+0.5*self.c_bottom,web_nodes[0][1]],
                            [-bp_bottom+self.c_bottom,web_nodes[0][1]]
                            ]
        
        for node in bottom_nodes:
            self.add(node,part='bottom_flange')
        
        
        # Corner points
        # flange-to-web corner, bottom
        x_bottom_web = (-rm,web_nodes[0][1]+rm)
        
        newPts = corner_points(x_bottom_web,rm,phi=[1.5*np.pi,2.0*np.pi],n=n)
        
        for newPt in newPts[1:-1]:            
            self.add(newPt)
        
        # Web points
        self.add([0,-0.5*bp_web+rm],part='web')
        self.add([0,-0.5*bp_web+rm+0.25*hweb],part='web')
        self.add([0,-0.5*bp_web+rm+0.75*hweb],part='web')
        self.add([0,0.5*bp_web-rm],part='web')
        
        # Corner points
        # web-to-flange corner, top
        x_top_web = (+rm,web_nodes[1][1]-rm)
        
        newPts = corner_points(x_top_web,rm,phi=[np.pi,0.5*np.pi],n=n)
        
        for newPt in newPts[1:-1]:            
            self.add(newPt)
        
        # Top flange
        top_nodes = [[rm,web_nodes[1][1]]]
        
        if bp_top_lip > 0.0:
            top_nodes += [[0.25*bp_top,web_nodes[1][1]],
                          [0.75*bp_top,web_nodes[1][1]],
                          [bp_top-rm,web_nodes[1][1]]]
            
            # Corner
            # web-to-flange corner, top
            x_top_lip = (bp_top-rm,web_nodes[1][1]-rm)
        
            for node in top_nodes:
                self.add(node,part='top_flange')
        
            newPts = corner_points(x_top_lip,rm,phi=[0.5*np.pi,0],n=n)
        
            for newPt in newPts[1:-1]:            
                self.add(newPt)
                
            # Top lip nodes
            
            self.add([bp_top,web_nodes[1][1]-rm],'top_lip')
            self.add([bp_top,web_nodes[1][1]-0.5*bp_top_lip],'top_lip')
            self.add([bp_top,web_nodes[1][1]-bp_top_lip],'top_lip')
        else:
            top_nodes += [[0.5*bp_top,web_nodes[1][1]],[bp_top,web_nodes[1][1]]]
        
            for node in top_nodes:
                self.add(node,part='top_flange')
                
    
    def create_model(self):
        """ Creates calculation model according to 
            EN 1993-1-3:2006, Annex C 
        
            Origin is located at the middle of the web
        """
        
        """
        bp_web = self.bp_w
        bp_bottom = self.bp_b
        bp_top = self.bp_t
        bp_bottom_lip = self.bp_b_lip
        bp_top_lip = self.bp_t_lip
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
        """
        
        self.prof = OpenProf(self._points,self.t)
        
        """ Set nodes and segments to parts """
        for i, node in enumerate(self.prof.nodes):
            x = node.coord
        
        """
        bp_web = self.hw_center
        bp_bottom = self.bb_center
        bp_top = self.bt_center
        bp_bottom_lip = self.bb_lip_center
        bp_top_lip = self.bt_lip_center
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
        
        if bp_bottom_lip > 0.0:
            bottom_nodes = [[-bp_bottom,web_nodes[0][1]+bp_bottom_lip],
                            [-bp_bottom,web_nodes[0][1]]
                            ]
        else:
            bottom_nodes = [[-bp_bottom,web_nodes[0][1]]]
            
        if bp_top_lip > 0.0:
            top_nodes = [[bp_top,web_nodes[1][1]],
                             [bp_top,web_nodes[1][1]-bp_top_lip]]
        else:
            top_nodes = [[bp_top,web_nodes[1][1]]]
        
        nodes = bottom_nodes + web_nodes + top_nodes
        
        self.prof = OpenProf(nodes,self.t)
        """
        
    def create_effective_model(self):
        """ Create calculation model for the effective cross-section 
            This is done based on the model of the gross cross-section
        """
    
        if self.prof is None:
            self.create_model()
                    
        prof_eff = deepcopy(self.prof)
        
        if self._rho_w < 1.0:
            """ Effective web """            
            if self._psi_w < 0.0:
                hc = self.hw_center/(1-self._psi_w)
                beff = self._rho_w * hc
                be1 = 0.4*beff
                be2 = 0.6*beff                
            elif self._psi_w < 1.0:
                hc = self.hw_center
                beff = self._rho_w * self.hw_center
                be1 = 2*beff/(5-self._psi_w)
                be2 = beff-be1            
            else:
                hc = self.hw_center
                beff = self._rho_w * self.hw_center
                be1 = 0.5*beff
                be2 = 0.5*beff
        
            bneg = hc-beff
            # Part of the web in tension
            ht = self.hw_center-hc
        
            if self.cb > 0:
                nweb = 2
            else:
                nweb = 1
            
            nw_top = prof_eff.parts['web']['nodes'][2]
            nw_bot = prof_eff.parts['web']['nodes'][1]
            
            
            prof_eff.nodes[nw_top].z = prof_eff.nodes[nw_top+1].z + self.gr - be1
            prof_eff.nodes[nw_bot].z = prof_eff.nodes[nw_top].z - bneg
            
            prof_eff.segments[nw_bot].t = 0.0
            
            #s = (be2+ht)/self.hw_center
            #prof_eff.split_segment(nweb,s)
            #s2 = bneg/(be1+bneg)
            #prof_eff.split_segment(nweb+1,s2)
            #prof_eff.segments[nweb+1].t = 0.0
        
        if self.ca > 0.0:            
            """ Edge stiffener is present """
            # Number of segments in the section            
            #nseg = prof_eff.n-1
            
            # Move the node of the edge stiffener up
            #prof_eff.nodes[-1].z += (1-self._rho_t_lip)*self.bt_lip_center
                    
            # Split the top flange into two parts
            #s = 0.5*self._rho_t            
            #prof_eff.split_segment(nseg-2,s)
                        
            #prof_eff.draw(node_labels=True)
            if self._rho_t < 1.0:
                s2 = (1-self._rho_t)*self.bt_center/prof_eff.segments[-2].length()
                prof_eff.split_segment(nseg-1,s2)
                prof_eff.segments[nseg-1].t = 0.0
            prof_eff.segments[-2].t *= self._chi_d_t
            prof_eff.segments[-1].t *= self._chi_d_t
                        
        else:
            """ The effective top flange is an outstand element """
            #prof_eff.nodes[-2].y -= (1-self._rho_t)*self.bp_t + self.gr
            prof_eff.nodes[-2].y = self._rho_t*self.bp_t + self.gr
            prof_eff.segments[-1].t = 0.0

        if self.cb > 0.0:
            """ Edge stiffener is present in the bottom flange """
            
            # Move the node of the edge stiffener down
            prof_eff.nodes[0].z -= (1-self._rho_b_lip)*self.bb_lip_center
            
            # Split the bottom flange into two parts
            s = 0.5*self._rho_b
            prof_eff.split_segment(1,s)
                        
            """ If there is a non-effective part of the bottom flange
                remove that too.
            """
            if self._rho_b < 1.0:
                s2 = (1-self._rho_b)*self.bb_center/prof_eff.segments[2].length()
                prof_eff.split_segment(2,s2)
                prof_eff.segments[2].khp = 0.0
            prof_eff.segments[0].t *= self._chi_d_b
            prof_eff.segments[1].t *= self._chi_d_b

        self.prof_eff = prof_eff
                            
    
    def create_edge_stiffener(self,flange='top'):
        """ Creates a model for the edge stiffener of a flange """
        if flange == 'top':
            n1 = self.parts['top_flange']['nodes'][-2]
            n2 = self.parts['top_lip']['nodes'][-1]
            nodes = self.prof.nodes[n1:n2+1]
            
            #corner = list(self.prof.nodes[-2].coord)
            #y_edge = corner[0]-0.5*self.bt_center*self._rho_t
            #z_lip  = corner[1]-self.bt_lip_center*self._rho_t_lip
            #nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
            
            self.edge_stiff[flange].draw()
        else:
            corner = list(self.prof.nodes[1].coord)            
            y_edge = corner[0]+0.5*self.bb_center*self._rho_b
            z_lip  = corner[1]+self.bb_lip_center*self._rho_b_lip
            nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
    
    def create_free_flange(self,flange='bottom'):
        """ Creates a model for the free flange
            This includes the bottom flange with edge stiffener and
            1/5 of the height of the web
        """
        if flange == 'top':            
            corner = list(self.prof.nodes[-2].coord)
            y_edge = corner[0]-0.5*self.bt_center*self._rho_t
            z_lip  = corner[1]-self.bt_lip_center*self._rho_t_lip
            nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
        else:
            nodes = []
            for i in range(0,3):
                nodes.append(list(self.prof.nodes[i].coord))
                        
            web_node = [0,nodes[-1][1]]
            web_node[1] += 0.2*self.hw 

            nodes.append(web_node)
            
            self.free_flange[flange] = OpenProf(nodes,self.t)
        
        
    
    @property
    def hw(self):
        """ Straight part of the web """
        return self.h - 2*(self.r+self.t_nom)
    
    @property
    def hw_center(self):
        """ Height of the profile from center line of the top flange
            to the center line of the bottom flange.
        """
        return self.h - self.t_nom
    
    @property
    def bt_center(self):
        """ width of the top flange either from the edge to the
            center line of the web or from center line of edge stiffener
            to the center line of the web
        """
        if self.ca == 0:
            bt = self.a - 0.5*self.t_nom
        else:
            bt = self.a - self.t_nom
        return bt
    
    @property
    def bb_center(self):
        """ width of the bottom flange either from the edge to the
            center line of the web or from center line of edge stiffener
            to the center line of the web
        """
        if self.cb == 0:
            bb = self.b - 0.5*self.t_nom
        else:
            bb = self.b - self.t_nom
        return bb
    
    @property
    def bt_lip_center(self):
        """ width of the top edge stiffener to the 
            to the center line of the top flange
        """
        if self.ca == 0:
            return 0
        else:
            return self.ca-0.5*self.t_nom
        
    @property
    def bb_lip_center(self):
        """ width of the bottom edge stiffener to the 
            to the center line of the bottom flange
        """
        if self.cb == 0:
            return 0
        else:
            return self.cb-0.5*self.t_nom
    
    @property
    def c_top(self):
        """ Straight part of the top flange """
        if self.ca > 0:
            c = self.a-2*(self.r+self.t_nom)
        else:
            c = self.a-(self.r+self.t_nom)
        return c
    
    @property
    def c_bottom(self):
        """ Straight part of the bottom flange """
        if self.cb > 0:
            c = self.b-2*(self.r+self.t_nom)
        else:
            c = self.b-(self.r+self.t_nom)
        return c
    
    @property
    def ca_straight(self):
        """ Straight part of the top lip """
        if self.ca > 0:
            c = self.ca-self.t_nom-self.r
        else:
            c = 0.0
        return c
    
    @property
    def cb_straight(self):
        """ Straight part of the bottom lip """
        if self.cb > 0:
            c = self.cb-self.t_nom-self.r
        else:
            c = 0.0
        
        return c
    
    @property
    def rm(self):
        """ Radius to the center line """
        return self.r + 0.5*self.t_nom
    
    def flange_lip(self,flange='top'):
        
        if flange == 'top':
            return self.ca
        else:
            return self.cb
    
    def center_line_length(self):    
        straight_parts = self.cb_straight + self.c_bottom + self.hw + self.c_top + self.ca_straight 
        
        n = 2
        if self.ca > 0.0:
            n += 1
        
        if self.cb > 0.0:
            n += 1
        
        return straight_parts +n*np.pi*self.rm/2
    
    def Agross(self):
        """ Gross cross-sectional area """
        
        return self.t * self.center_line_length()
    
    def effective_width_flange(self,psi,flange='top',verb=False):
        """ Effective width of a flange """
        if flange == "top":
            bp = self.bp_t
            edge_stiffener = abs(self.ca) > 1e-10
        else:
            bp = self.bp_b
            edge_stiffener = abs(self.cb) > 1e-10
        
        if edge_stiffener:
            ksigma = en1993_1_5.buckling_factor_internal(psi)
            lambda_p = en1993_1_5.lambda_p(bp,self.t,self.material.eps(),ksigma)
            rho = en1993_1_5.reduction_factor_internal(lambda_p, psi)
            beff, be = en1993_1_5.effective_width_internal(bp,psi,rho)
            
        else:
            ksigma = en1993_1_5.buckling_factor_outstand(psi)
            lambda_p = en1993_1_5.lambda_p(bp,self.t,self.material.eps(),ksigma)
            rho = en1993_1_5.reduction_factor_outstand(lambda_p)
            beff = en1993_1_5.effective_width_outstand(bp,psi,rho)
            
        if flange == 'top':
            self._rho_t = rho
            self._lambda_t = lambda_p
        else:
            self._rho_b = rho
            self._lambda_b = lambda_p
        
        if verb:
            print("** Effective {0:s} flange **".format(flange))
            print("  Notional width: bp = {0:4.2f}".format(bp))
            print("  Buckling factor: ksigma = {0:4.3f}".format(ksigma))
            print("  Slenderness: lambda_p = {0:4.3f}".format(lambda_p))
            print("  Reduction factor: rho = {0:4.3f}".format(rho))
            print("  Effective width: beff = {0:4.3f} mm".format(beff))
            
    def effective_width_edge_stiffener(self,flange='top',verb=False):
        """ Effective width of an edge stiffener """
        if flange == "top":
            bp = self.bp_t
            bpc = self.bp_t_lip
        else:
            bp = self.bp_b
            bpc = self.bp_b_lip
        
        # NOTE: Buckling factor according to EN 1993-1-3, not EN 1993-1-5!
        ksigma = en1993_1_3.buckling_factor_edge_stiffener(bpc/bp)
        lambda_p = en1993_1_5.lambda_p(bpc,self.t,self.material.eps(),ksigma)
        rho = en1993_1_5.reduction_factor_outstand(lambda_p)
        ceff = en1993_1_5.effective_width_outstand(bpc,psi=1.0,rho=rho)
            
        if flange == 'top':
            self._rho_t_lip = rho
            self._lambda_t_lip = lambda_p
        else:
            self._rho_b_lip = rho
            self._lambda_b_lip = lambda_p
            
        if verb:
            print("** Effective {0:s} edge stiffener **".format(flange))
            print("  Notional width of flange: bp = {0:4.2f}".format(bp))
            print("  Notional width of stiffener: bp,c = {0:4.2f}".format(bpc))
            print("  Buckling factor: ksigma = {0:4.3f}".format(ksigma))
            print("  Slenderness: lambda_p = {0:4.3f}".format(lambda_p))
            print("  Reduction factor: rho = {0:4.3f}".format(rho))
            print("  Effective width: ceff = {0:4.3f} mm".format(ceff))
        
        
    
    def effective_width_web(self,psi,verb=False):
        """ Effective width of web """
        bp = self.bp_w
        ksigma = en1993_1_5.buckling_factor_internal(psi)        
        lambda_p = en1993_1_5.lambda_p(bp,self.t,self.material.eps(),ksigma)
        rho = en1993_1_5.reduction_factor_internal(lambda_p, psi)
        beff, be = en1993_1_5.effective_width_internal(bp,psi,rho)
        
        if verb:
            print("** Effective web **")
            print("  Notional width: bp = {0:4.2f} mm".format(bp))
            print("  Stress ratio: psi = {0:4.3f}".format(psi))
            print("  Buckling factor: ksigma = {0:4.3f}".format(ksigma))
            print("  Slenderness: lambda_p = {0:4.3f}".format(lambda_p))
            print("  Reduction factor: rho = {0:4.3f}".format(rho))
            print("  Effective width: beff = {0:4.3f} mm".format(beff))
    
        self._rho_w = rho
        self._psi_w = psi
        self._lambda_w = lambda_p
    
    def distortional_buckling(self,flange='top',verb=False):
        """ Distortional buckling of the edge stiffene 
            1. Generate edge stiffener effective part, As, Is
            2. Calculate stiffness K
            3. Calculate critical stress
            4. Calculate slenderness
            5. Calculate reduction factor for thickness reduction.
        
        """        
        K = self.edge_stiffness(flange)
        Is = self.edge_stiff[flange].second_moments_gc()[0]
        As = self.edge_stiff[flange].area()
        E = self.material.E
        
        scr = en1993_1_3.distortional_buckling_stress(Is,As,E,K)
        lambda_d = np.sqrt(self.fyb/scr)
        chi_d = en1993_1_3.distortional_buckling_reduction_factor(lambda_d)
        if verb:
            print("Distortional buckling:")
            print("  K = {0:4.3f} N/mm2".format(K))
            print("  s_cr = {0:4.3f} N/mm2".format(scr))
            print("  lambda_d = {0:4.3f}".format(lambda_d))
            print("  chi_d = {0:4.3f}".format(chi_d))
        
        if flange == 'top':
            self._chi_d_t = chi_d
            self._lambda_t = lambda_d
        else:
            self._chi_d_b = chi_d
            self._lambda_b = lambda_d
            
        return chi_d
    
    def area(self,delta=False):
        """ Gross cross-sectional area 
        
            :param : delta .. Flag for stating whether or not
                    the Eq. (5.1) or EN 1993-1-3 is used or not.
        """
        
        if delta:
            r = [self.r,self.r]
            phi = [90,90]
            
            bp = [self.bp_b, self.bp_w, self.bp_t]
            
            if self.ca > 0.0:
                r.append(self.r)
                phi.append(90)
                bp.append(self.bp_t_lip)
            
            if self.cb > 0.0:
                r.append(self.r)
                phi.append(90)
                bp.append(self.bp_b_lip)
            
            
            d = en1993_1_3.rounding_factor(r,phi,bp)
        else:
            d = 0
        
        return self.prof.area()*(1-d)
    
    def centroid(self):
        
        ygc, zgc, A = self.prof.centroid()
        
        return ygc, zgc
    
    def centroid_eff(self):
        """ Centroid of the effective cross-section """
        
        ygc, zgc, A = self.prof_eff.centroid()
        
        return ygc, zgc
    
    def second_moment(self):
        """ second moment of area for the gross cross section 
            with respect to 
        """
    
    def effective_section(self,load='compression',verb=False):
        """ Returns effective section, including local buckling
            and distortional buckling.
            
            The effective section contains:
                _rho_t .. effective top flange
                _rho_w .. effective web
                _rho_b .. effective bottom flange
                _rho_t_lip .. effective top edge stiffener
                _rho_t_bottom .. effective bottom edge stiffener
                
                _chid_t .. reduction factor for top edge stiffener for
                            distortional buckling
                _chid_b .. reduction factor for bottom edge stiffener
                            for distortional buckling
        """
        
        if load == 'bending_y_pos':
            """ Positive bending:
                top flange in compression
                bottom flange in tension
            """
            self.effective_width_flange(psi=1.0,flange='top',verb=verb)
            
            if self.ca > 0:
                self.effective_width_edge_stiffener(flange='top',verb=verb)                
                self.create_edge_stiffener(flange='top')
                self.distortional_buckling(flange='top',verb=verb)
            
            
            """ Determine stress ratio for the web, using effecive flange """
            self.create_effective_model()
            
            self.prof_eff.draw(coord_axes=True)
            psi = self.stress_ratio()            
                        
            if verb:
                print("Psi = {0:4.3f}".format(psi))
            
            self.effective_width_web(psi,verb=verb)
            self.create_effective_model()
        elif load == 'compression':
            """ Compression: all parts are checked """
            for flange in ['top','bottom']:
                self.effective_width_flange(psi=1.0,flange=flange,verb=verb)
                if self.flange_lip(flange) > 0:
                    self.effective_width_edge_stiffener(flange,verb)
                    self.create_edge_stiffener(flange)
                
                    
                    
            for flange in ['top','bottom']:
                if self.flange_lip(flange) > 0:
                    self.distortional_buckling(flange=flange,verb=verb)
            
            
            self.effective_width_web(psi=1.0,verb=verb)            
            self.create_effective_model()
            
    def stress_ratio(self):
        """ Calculates the ratio of the top flange stress and
            bottom flange stress
        """
        
        if self.prof_eff is None:
            ygc, zgc = self.centroid()
            print("No effective model created")
        else:
            ygc, zgc = self.centroid_eff()
            
        """ Height of the compressed part of the profile
            NOTE:
                zgc is determined using the simplified computational model,
                where the height of the profile is smaller than 'h', namely
                the thickness of the profile is reduced as notional widths
                are employed.
                
                Maybe the height of the computational profile should be
                used instead of 'h'
        """
        
        # Use hw_center to go from center line to center line.
        hc = 0.5*self.hw_center - zgc
        
        print("Centroid: ygc = {0:4.3f}, zgc = {1:4.3f}".format(ygc,zgc))
        print("Compressed part of the profile: {0:4.2f} mm".format(hc))
        
        return -(self.hw_center/hc-1)

    def edge_stiffness(self,flange='top'):
        """ EN 1993-1-3, Eq. (5.10b)
            Stiffness provided by the profile to the edge stiffener
        """
        
        C = 0.25*self.material.E*self.t**3/(1-self.material.nu**2)        
        hw = self.hw_center
        
        """ TODO:
            Implement b1 and b2 (distance from far edge of the flange to
            the centroid of effective stiffener)
        """
        
        if flange == 'top':
            f1 = 'top'
            f2 = 'bottom'
        else:
            f1 = 'bottom'
            f2 = 'top'
        
        if self.Ned < 0.0:
            if self.issymmetric():
                kf = 1.0
            else:
                kf = self.edge_stiff[f2].area()/self.edge_stiff[f1].area()
        elif self.Med > 0.0 and flange == 'top':            
            kf = 0.0
        elif self.Med < 0.0 and flange == 'bottom':
            kf = 0.0
        else:
            kf = 0.0
        
        b1 = abs(self.edge_stiff[f1].centroid()[0])
        #print('b1 = {0:4.3f}'.format(b1))
        if kf != 0.0:
            b2 = abs(self.edge_stiff[f2].centroid()[0])
        else:
            b2 = 0.0
        
        print(f'kf = {kf:4.3f}')
        print(f'b1 = {b1:8.7f}')
        print(f'b2 = {b2:8.7f}')
        print(f'hw = {hw:6.5f}')
        print(f'C = {C:6.5f}')
        print(f'den = {b1**2*hw + b1**3 + 0.5*b1*b2*hw*kf:6.5f}')
        print(f'K = {C/(b1**2*hw + b1**3 + 0.5*b1*b2*hw*kf):6.5f}')
        
        return C/(b1**2*hw + b1**3 + 0.5*b1*b2*hw*kf)
    
    def kh(self,load='gravity',a=0.0):
        """ Equivalent lateral load for purlins """
    
        xs = self.shear_center
        e = xs[0]
        h = self.h
        gs = 0.5*h-xs[1]
    
        kh0 = self.Iyz/self.Iy * gs/h        
        
        if load == 'gravity':
            kh = kh0+e/h
            if e < 0:
                """ shear center is left of qEd:
                    load is to the left, i.e. in the negative
                    direction
                """
                kh *= -1.0                
        elif load == 'uplift':
            aS = a-e
            kh = kh0+aS/h
            
            if aS/h >= kh0:
                kh *= -1.0                

        return kh
    
    def KB(self,load='gravity',afp=0.0):
        """ Stiffness of purlin-sheeting connection """
        
        if afp == 0.0:
            """ If no fastener position is given, use half of the width
                of the top flange
            """
            afp = 0.5*self.a
        
        if self.kh(load) > 0:
            bmod = afp
        else:
            bmod = 2*afp + self.b
            
        hd = self.hw_center
                
        return 0.25*self.material.E*self.t**3/(self.hw_center**2*(1-self.material.nu**2)*(hd+bmod))
        
    def K(self,load='gravity',afp=0.0,CD=0.0):
        """ Stiffness of the purlin-sheeting connection """
        KB = self.KB(load,afp)
        
        if CD != 0.0:
            KA_inv = self.hw_center**2/CD
        else:
            ba = self.a
            tnom = 1.0
            position = 'pos'
            A = 10
            bT = 50
            bR = 150
            
            CD = en1993_1_3.CD_A(ba,tnom,position,A,bT,bR,load)
            print("C_D = {0:4.3f}".format(CD))
            KA_inv = self.hw_center**2/CD
        
        print("KB = {0:4.3f}".format(KB))
        print("KA = {0:4.3f}".format(1/KA_inv))
        
        return 1/(1/KB+KA_inv)

    def axial_force_resistance(self,verb=False):
        if self.Ned >= 0:
            NRd = self.fya*self.A/gammaM0
        else:
            Aeff = self.Aeff
            Ag = self.A
            if Aeff < Ag:
                NRd = self.fyb*Aeff/gammaM0
            else:
                lambda_ratio = self._lambda_w/0.673
                
                if self.ca > 0.0:
                    top_lambda_ratio = self._lambda_t/0.65
                else:
                    top_lambda_ratio = self._lambda_t/0.673
                
                if self.cb > 0.0:
                    bot_lambda_ratio = self._lambda_b/0.65
                else:
                    bot_lambda_ratio = self._lambda_b/0.673
                    
                lambda_ratio = max(lambda_ratio,top_lambda_ratio,bot_lambda_ratio)
                NcRd = Ag*(self.fyb+4*(self.fya-self.fyb)*(1-lambda_ratio))
                
                NRd = min(NcRd,self.fya*Ag)/gammaM0

        if verb:
            print("NRd = {0:4.2f} kN".format(NRd*1e-3))

        return NRd

    def export_cufsm(self,n=4):
        """ Exports the cross-section to be analyzed with CUFSM5 software
            run in Matlab            
        """
        print("Material:")
        print(f"100 {self.material.E:7.2f} {self.material.E:7.2f} {self.material.nu:3.1f} {self.material.nu:3.1f} {self.material.G:7.2f}")
        
        
        

    def draw(self):
        """ Draw the profile 
            
        """        
        
        fig, ax = plt.subplots(1)        
        
        lw = 1.5    # Line width of the outer line
        lw_c = 1.25 # Line width of the center line
        col = 'b'
        
        h = self.h
        hweb = self.hw
        b_top = self.a
        b_bot = self.b
        ca = self.ca
        cb = self.cb
        tnom = self.t_nom
        r = self.r
        
        # Draw webs
        y_web = [-0.5*hweb,0.5*hweb]                            
        ax.vlines([-0.5*tnom,0.5*tnom],y_web[0],y_web[1],colors='k',linewidth=lw)
        ax.vlines([0.0],y_web[0],y_web[1],colors='b',linewidth=lw_c,linestyle='-.')
        
        
        # Draw top flange
        ctop = self.c_top # Straight part of the top flange
        x_top = [0.5*tnom+r,0.5*tnom+r+ctop]
        y_top = [0.5*h,0.5*h-tnom]
        
        ax.hlines(y_top,x_top[0],x_top[1],colors='k',linewidth=lw)
        ax.hlines([0.5*(h-tnom)],x_top[0],x_top[1],colors='b',linewidth=lw_c,linestyle='-.')
    
        
        # Corner from web to top flange
        x_web = (0.5*tnom+r,0.5*h-tnom-r)
        
        up_corner_in = patches.Arc(x_web,2*r,2*r,0,90,180,color='k',linewidth=lw)
        up_corner_out = patches.Arc(x_web,2*(r+tnom),2*(r+tnom),0,90,180,color='k',linewidth=lw)
        up_corner_mid = patches.Arc(x_web,2*r+tnom,2*r+tnom,0,90,180,color='b',linewidth=lw_c,linestyle='-.')
        
        ax.add_patch(up_corner_in)
        ax.add_patch(up_corner_out)
        ax.add_patch(up_corner_mid)
        
        # Draw top lip
        if self.ca > 0.0:
            y_top_lip = [0.5*hweb-(ca-tnom-r),0.5*hweb]                            
            ax.vlines([b_top-1.5*tnom,b_top-0.5*tnom],y_top_lip[0],y_top_lip[1],colors='k',linewidth=lw)
            ax.vlines([b_top-tnom],y_top_lip[0],y_top_lip[1],colors='b',linewidth=lw_c,linestyle='-.')
        
            # Corner of top lip
            x_top_right = (b_top-1.5*tnom-r,0.5*h-tnom-r)
        
            top_right_corner_in = patches.Arc(x_top_right,2*r,2*r,0,0,90,color='k',linewidth=lw)
            top_right_corner_out = patches.Arc(x_top_right,2*(r+tnom),2*(r+tnom),0,0,90,color='k',linewidth=lw)
        
            ax.add_patch(top_right_corner_in)
            ax.add_patch(top_right_corner_out)
        
            ax.hlines(0.5*hweb-(ca-tnom-r),b_top-1.5*tnom,b_bot-0.5*tnom,colors='k',linewidth=lw)
        
        # Bottom flange
        cbottom = self.c_bottom # Straight part of the top flange
        x_bot = [-(0.5*tnom+r),-(0.5*tnom+r+cbottom)]
        y_bot = [-(0.5*h-tnom),-0.5*h]
        
        ax.hlines(y_bot,x_bot[0],x_bot[1],colors='k',linewidth=lw)
        ax.hlines([-0.5*(h-tnom)],x_bot[0],x_bot[1],colors='b',linewidth=lw_c,linestyle='-.')
        
        # Corner from web to bottom flange
        x_web = (-(0.5*tnom+r),-(0.5*h-tnom-r))
        
        bot_corner_in = patches.Arc(x_web,2*r,2*r,0,270,360,color='k',linewidth=lw)
        bot_corner_out = patches.Arc(x_web,2*(r+tnom),2*(r+tnom),0,270,360,color='k',linewidth=lw)
        bot_corner_mid = patches.Arc(x_web,2*r+tnom,2*r+tnom,0,270,360,color='b',linewidth=lw_c,linestyle='-.')
        
        ax.add_patch(bot_corner_in)
        ax.add_patch(bot_corner_out)
        ax.add_patch(bot_corner_mid)
        
        # Left lip      
        if self.cb > 0.0:
            y_bot_lip = [-(0.5*hweb-(cb-tnom-r)),-(0.5*hweb)]                            
            ax.vlines([-(b_bot-1.5*tnom),-(b_bot-0.5*tnom)],y_bot_lip[0],y_bot_lip[1],colors='k',linewidth=lw)
            ax.vlines([-(b_bot-tnom)],y_bot_lip[0],y_bot_lip[1],colors='b',linewidth=lw_c,linestyle='-.')
            
            # Corner of the lip
            x_bot_left = (-(b_bot-1.5*tnom-r),-(0.5*h-tnom-r))
            
            bot_left_corner_in = patches.Arc(x_bot_left,2*r,2*r,0,180,270,color='k',linewidth=lw)
            bot_left_corner_out = patches.Arc(x_bot_left,2*(r+tnom),2*(r+tnom),0,180,270,color='k',linewidth=lw)
            bot_left_corner_mid = patches.Arc(x_bot_left,2*r+tnom,2*r+tnom,0,180,270,color='b',linewidth=lw_c,linestyle='-.')
            
            ax.add_patch(bot_left_corner_in)
            ax.add_patch(bot_left_corner_out)
            ax.add_patch(bot_left_corner_mid)
            
            ax.hlines(-(0.5*hweb-(cb-tnom-r)),-b_bot+0.5*tnom,-b_bot+1.5*tnom,colors='k',linewidth=lw)
        
        
        
        ax.set_aspect('equal')
        
class CSection(ZSection):
    """ C section
        This profile has many things in common with the Z section,
        so it inherits the basic functionalities
    """
    
    def __init__(self,t_nom=2.0,h=120,a=60,b=60,ca=20,cb=20,r=2.0,material="S350GD",t_coat=0.04):
        
        super().__init__(t_nom,h,a,b,ca,cb,r,material,t_coat)
    
    def kh0(self):
        """ Coefficient kh0 from Fig. 10.3 """
        sc = self.shear_center
        gs = 0.5*self.h-sc[1]
        kh0 = self.Iyz/self.Iy*gs/self.h
        
        return kh0, sc[0]
        
    def kh(self,load='down'):
        """ Coefficient kh from Fig. 10.3 """
        
        kh0, y_sc = self.kh0()
        
        if load == 'down':
            kh = kh0 + y_sc/self.h
        else:
            f = 0.5*self.bt_center - y_sc
            kh = kh0 - f/self.h
            
        return kh
    
    def create_points(self,n=5):
        """ Creates points for the calculation model 
            :param n: number of divisions in the corner areas
            
        """
        # Number of points; used as an index
        npt = 0
        
        self.corner_pts = n
        h = self.h
        hweb = self.hw
        
        tnom = self.t_nom
        rm = self.Rm
        ro = self.Ro
        r = self.r
        
        bp_web = self.hw_center
        bp_bottom = self.bb_center
        bp_top = self.bt_center
        bp_bottom_lip = self.bb_lip_center
        bp_top_lip = self.bt_lip_center
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
        
        if bp_bottom_lip > 0.0:
            bottom_lip_nodes = [[bp_bottom,web_nodes[0][1]+bp_bottom_lip],
                                [bp_bottom,web_nodes[0][1]+bp_bottom_lip-0.5*self.cb_straight]]
            print(bottom_lip_nodes)
            for node in bottom_lip_nodes:
                self.add(node,part='bottom_lip')
            
            # Corner points
            # Lip-to-flange corner, bottom
            x_bottom_lip = (bp_bottom-rm,web_nodes[0][1]+rm)
        
            newPts = corner_points(x_bottom_lip,rm,phi=[0,-0.5*np.pi],n=n)
            
            self.add(newPts[0],part='bottom_lip')
            
            for newPt in newPts[1:-1]:            
                self.add(newPt)
            
            # The second and third nodes are for helping to cope
            # with the effective cross section.
            bottom_nodes = [[bp_bottom-rm,web_nodes[0][1]],
                            [bp_bottom-rm-1/3*self.c_bottom,web_nodes[0][1]],
                            [bp_bottom-rm-2/3*self.c_bottom,web_nodes[0][1]],
                            [bp_bottom-rm-self.c_bottom,web_nodes[0][1]]
                            ]            
        else:
            bottom_nodes = [[bp_bottom,web_nodes[0][1]],
                            [bp_bottom-0.5*self.c_bottom,web_nodes[0][1]],
                            [bp_bottom-self.c_bottom,web_nodes[0][1]]
                            ]
        
        for node in bottom_nodes:
            self.add(node,part='bottom_flange')
        
        
        # Corner points
        # flange-to-web corner, bottom
        x_bottom_web = (rm,web_nodes[0][1]+rm)
        
        newPts = corner_points(x_bottom_web,rm,phi=[1.5*np.pi,np.pi],n=n)
        
        for newPt in newPts[1:-1]:            
            self.add(newPt)
        
        # Web points
        self.add([0,-0.5*bp_web+rm],part='web')
        self.add([0,-0.5*bp_web+rm+0.25*hweb],part='web')
        self.add([0,-0.5*bp_web+rm+0.75*hweb],part='web')
        self.add([0,0.5*bp_web-rm],part='web')
        
        # Corner points
        # web-to-flange corner, top
        x_top_web = (+rm,web_nodes[1][1]-rm)
        
        newPts = corner_points(x_top_web,rm,phi=[np.pi,0.5*np.pi],n=n)
        
        for newPt in newPts[1:-1]:            
            self.add(newPt)
        
        # Top flange
        top_nodes = [[rm,web_nodes[1][1]]]
        
        if bp_top_lip > 0.0:
            top_nodes += [[0.25*bp_top,web_nodes[1][1]],
                          [0.75*bp_top,web_nodes[1][1]],
                          [bp_top-rm,web_nodes[1][1]]]
            
            # Corner
            # web-to-flange corner, top
            x_top_lip = (bp_top-rm,web_nodes[1][1]-rm)
        
            for node in top_nodes:
                self.add(node,part='top_flange')
        
            newPts = corner_points(x_top_lip,rm,phi=[0.5*np.pi,0],n=n)
        
            for newPt in newPts[1:-1]:            
                self.add(newPt)
                
            # Top lip nodes
            
            self.add([bp_top,web_nodes[1][1]-rm],'top_lip')
            self.add([bp_top,web_nodes[1][1]-0.5*bp_top_lip],'top_lip')
            self.add([bp_top,web_nodes[1][1]-bp_top_lip],'top_lip')
        else:
            top_nodes += [[0.5*bp_top,web_nodes[1][1]],[bp_top,web_nodes[1][1]]]
        
            for node in top_nodes:
                self.add(node,part='top_flange')
                
    
    #def create_model(self):
        """ Creates calculation model according to 
            EN 1993-1-3:2006, Annex C 
        
            Origin is located at the middle of the web
        """
        """
        bp_web = self.bp_w
        bp_bottom = self.bp_b
        bp_top = self.bp_t
        bp_bottom_lip = self.bp_b_lip
        bp_top_lip = self.bp_t_lip
        """
        
        """
        bp_web = self.hw_center
        bp_bottom = self.bb_center
        bp_top = self.bt_center
        bp_bottom_lip = self.bb_lip_center
        bp_top_lip = self.bt_lip_center
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
        if bp_bottom_lip > 0.0:
            bottom_nodes = [[bp_bottom,web_nodes[0][1]+bp_bottom_lip],
                            [bp_bottom,web_nodes[0][1]]
                            ]
        else:
            # No bottom edge stiffener
            bottom_nodes = [[bp_bottom,web_nodes[0][1]]]
        
        if bp_top_lip > 0.0:
            top_nodes = [[bp_top,web_nodes[1][1]],
                             [bp_top,web_nodes[1][1]-bp_top_lip]]
        else:
            top_nodes = [[bp_top,web_nodes[1][1]]]
            
        nodes = bottom_nodes + web_nodes + top_nodes
        
        self.prof = OpenProf(nodes,self.t)
        """
    
    def create_effective_model(self):
        """ Create calculation model for the effective cross-section 
            This is done based on the model of the gross cross-section
        """
    
        if self.prof is None:
            self.create_model()
                    
        prof_eff = deepcopy(self.prof)
                
        if self._rho_w < 1.0:
            """ Effective web """            
            if self._psi_w < 0.0:
                hc = self.bp_w/(1-self._psi_w)
                beff = self._rho_w * hc
                be1 = 0.4*beff
                be2 = 0.6*beff                
            elif self._psi_w < 1.0:
                hc = self.bp_w
                beff = self._rho_w * hc
                be1 = 2*beff/(5-self._psi_w)
                be2 = beff-be1            
            else:
                hc = self.bp_w
                beff = self._rho_w * hc
                be1 = 0.5*beff
                be2 = 0.5*beff
        
            bneg = hc-beff
            # Part of the web in tension
            #ht = self.hw_center-hc
        
            if self.cb > 0:
                nweb = 2
            else:
                nweb = 1
            
            """
            s = (be2+ht)/self.hw_center
            prof_eff.split_segment(nweb,s)
            s2 = bneg/(be1+bneg)
            prof_eff.split_segment(nweb+1,s2)
            prof_eff.segments[nweb+1].t = 0.0
            """
            nw_top = self.parts['web']['nodes'][2]
            nw_bot = self.parts['web']['nodes'][1]
            
            print(be1,be2, bneg)
                    
            prof_eff.nodes[nw_top].z = prof_eff.nodes[nw_top+1].z + self.rm*np.sin(np.radians(45)) - be1
            prof_eff.nodes[nw_bot].z = prof_eff.nodes[nw_top].z - bneg
            
            prof_eff.segments[nw_bot].t = 0.0
        
        if self.ca > 0.0:            
            """ Edge stiffener is present """
            # Number of segments in the section
            
            ntop = self.parts['top_flange']['nodes'][-2]
            
            # Reduce the thickness of edge stiffener, if needed
            for segment in prof_eff.segments[ntop:]:
                segment.t *= self._chi_d_t
                
            # If the top flange is not entirely effective,
            # make the thickness of the non-effective segment zero.
            if self._rho_t < 1.0:
                prof_eff.segments[ntop-1].t = 0
            
                              
            
            """
            nseg = prof_eff.n-1
            # Move the node of the edge stiffener up
            prof_eff.nodes[-1].z += (1-self._rho_t_lip)*self.bt_lip_center
                    
            # Split the top flange into two parts
            s = 0.5*self._rho_t            
            prof_eff.split_segment(nseg-2,s)
                        
            #prof_eff.draw(node_labels=True)
            if self._rho_t < 1.0:
                s2 = (1-self._rho_t)*self.bt_center/prof_eff.segments[-2].length()
                prof_eff.split_segment(nseg-1,s2)
                prof_eff.segments[nseg-1].khp = 0.0
                
            prof_eff.segments[-2].t *= self._chi_d_t
            prof_eff.segments[-1].t *= self._chi_d_t
            """          
        else:
            """ The effective top flange is an outstand element """
            #prof_eff.nodes[-1].y -= (1-self._rho_t)*self.bp_t
            if self._rho_t < 1.0:
                prof_eff.nodes[-2].y = self._rho_t*self.bp_t + self.gr
                prof_eff.segments[-1].t = 0.0

        
        if self.cb > 0.0:
            """ Edge stiffener is present in the bottom flange """
            
            nbot = self.parts['bottom_flange']['nodes'][1]
            
            # Reduce the thickness of edge stiffener, if needed
            for segment in prof_eff.segments[:nbot]:
                segment.t *= self._chi_d_b
            
            # If the bottom flange is not entirely effective,
            # make the thickness of the non-effective segment zero.
            if self._rho_b < 1.0:
                prof_eff.segments[nbot].t = 0
            
            """
            # Move the node of the edge stiffener down
            prof_eff.nodes[0].z -= (1-self._rho_b_lip)*self.bb_lip_center
            
            # Split the bottom flange into two parts
            s = 0.5*self._rho_b
            prof_eff.split_segment(1,s)
            """            
            """ If there is a non-effective part of the bottom flange
                remove that too.
            """
            """
            if self._rho_b < 1.0:
                s2 = (1-self._rho_b)*self.bb_center/prof_eff.segments[2].length()
                prof_eff.split_segment(2,s2)
                prof_eff.segments[2].khp = 0.0
            prof_eff.segments[0].t *= self._chi_d_b
            prof_eff.segments[1].t *= self._chi_d_b        
            """
        else:
            
            """ The effective bottom flange is an outstand element """
            #prof_eff.nodes[-1].y -= (1-self._rho_t)*self.bp_t
            if self._rho_b < 1.0:
                prof_eff.nodes[1].y = self._rho_b*self.bp_b + self.gr
                
            

        self.prof_eff = prof_eff
            
    def create_edge_stiffener(self,flange='top'):
        """ Creates a model for the edge stiffener of a flange """
        if flange == 'top':            
            n1 = self.parts['top_flange']['nodes'][-2]
            n2 = self.parts['top_lip']['nodes'][-1]
            nodes = self.prof.nodes[n1:n2+1]
            
            nodes[0].y = self.bp_t+self.gr-0.5*self.bp_t*self._rho_t
            
            if self._rho_t_lip < 1.0:            
                nodes[-2].z = nodes[-1].z + (1-self._rho_t_lip)*self.bp_t_lip
            
            #corner = list(self.prof.nodes[-2].coord)
            #y_edge = corner[0]-0.5*self.bt_center*self._rho_t
            #z_lip  = corner[1]-self.bt_lip_center*self._rho_t_lip
            #nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
            #self.edge_stiff[flange].draw()
            """
            self.edge_stiff[flange].draw()
            corner = list(self.prof.nodes[-2].coord)
            y_edge = corner[0]-0.5*self.bt_center*self._rho_t
            z_lip  = corner[1]-self.bt_lip_center*self._rho_t_lip
            nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
            """
        else:
            n1 = self.parts['bottom_lip']['nodes'][0]
            n2 = self.parts['bottom_flange']['nodes'][1]            
            nodes = self.prof.nodes[n1:n2+1]
            
            nodes[-1].y = self.bp_b+self.gr-0.5*self.bp_b*self._rho_b
            
            if self._rho_b_lip < 1.0:
                nodes[1].z = nodes[0].z + (1-self._rho_b_lip)*self.bp_b_lip
            
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
            """
            corner = list(self.prof.nodes[1].coord)            
            y_edge = corner[0]-0.5*self.bb_center*self._rho_b
            z_lip  = corner[1]+self.bb_lip_center*self._rho_b_lip
            nodes = [[y_edge,corner[1]],corner,[corner[0],z_lip]]
            self.edge_stiff[flange] = OpenProf(nodes,self.t)
            """
    def draw(self):
        """ Draw the profile 
            
        """        
        
        fig, ax = plt.subplots(1)        
        
        lw = 1.5    # Line width of the outer line
        lw_c = 1.25 # Line width of the center line
        col = 'b'
        
        h = self.h
        hweb = self.hw
        b_top = self.a
        b_bot = self.b
        ca = self.ca
        cb = self.cb
        tnom = self.t_nom
        r = self.r
        
        # Draw webs
        y_web = [-0.5*hweb,0.5*hweb]                            
        ax.vlines([-0.5*tnom,0.5*tnom],y_web[0],y_web[1],colors='k',linewidth=lw)
        ax.vlines([0.0],y_web[0],y_web[1],colors='b',linewidth=lw_c,linestyle='-.')
        
        
        # Draw top flange
        ctop = self.c_top # Straight part of the top flange
        x_top = [0.5*tnom+r,0.5*tnom+r+ctop]
        y_top = [0.5*h,0.5*h-tnom]
        
        ax.hlines(y_top,x_top[0],x_top[1],colors='k',linewidth=lw)
        ax.hlines([0.5*(h-tnom)],x_top[0],x_top[1],colors='b',linewidth=lw_c,linestyle='-.')
        
        # Corner from web to top flange
        x_web = (0.5*tnom+r,0.5*h-tnom-r)
        
        up_corner_in = patches.Arc(x_web,2*r,2*r,0,90,180,color='k',linewidth=lw)
        up_corner_out = patches.Arc(x_web,2*(r+tnom),2*(r+tnom),0,90,180,color='k',linewidth=lw)
        up_corner_mid = patches.Arc(x_web,2*r+tnom,2*r+tnom,0,90,180,color='b',linewidth=lw_c,linestyle='-.')
        
        ax.add_patch(up_corner_in)
        ax.add_patch(up_corner_out)
        ax.add_patch(up_corner_mid)
        
        # Draw top lip
        if self.ca > 0.0:
            y_top_lip = [0.5*hweb-(ca-tnom-r),0.5*hweb]                            
            ax.vlines([b_top-1.5*tnom,b_top-0.5*tnom],y_top_lip[0],y_top_lip[1],colors='k',linewidth=lw)
            ax.vlines([b_top-tnom],y_top_lip[0],y_top_lip[1],colors='b',linewidth=lw_c,linestyle='-.')
            
            # Corner of top lip
            x_top_right = (b_top-1.5*tnom-r,0.5*h-tnom-r)
            
            top_right_corner_in = patches.Arc(x_top_right,2*r,2*r,0,0,90,color='k',linewidth=lw)
            top_right_corner_out = patches.Arc(x_top_right,2*(r+tnom),2*(r+tnom),0,0,90,color='k',linewidth=lw)
            
            ax.add_patch(top_right_corner_in)
            ax.add_patch(top_right_corner_out)
            
            ax.hlines(0.5*hweb-(ca-tnom-r),b_top-1.5*tnom,b_top-0.5*tnom,colors='k',linewidth=lw)
        
        # Bottom flange
        cbottom = self.c_bottom # Straight part of the top flange
        x_bot = [0.5*tnom+r,0.5*tnom+r+cbottom]
        y_bot = [-(0.5*h-tnom),-0.5*h]
        
        ax.hlines(y_bot,x_bot[0],x_bot[1],colors='k',linewidth=lw)
        ax.hlines([-0.5*(h-tnom)],x_bot[0],x_bot[1],colors='b',linewidth=lw_c,linestyle='-.')
        
        # Corner from web to bottom flange
        x_web = (0.5*tnom+r,-(0.5*h-tnom-r))
        
        bot_corner_in = patches.Arc(x_web,2*r,2*r,0,180,270,color='k',linewidth=lw)
        bot_corner_out = patches.Arc(x_web,2*(r+tnom),2*(r+tnom),0,180,270,color='k',linewidth=lw)
        bot_corner_mid = patches.Arc(x_web,2*r+tnom,2*r+tnom,0,180,270,color='b',linewidth=lw_c,linestyle='-.')
        
        ax.add_patch(bot_corner_in)
        ax.add_patch(bot_corner_out)
        ax.add_patch(bot_corner_mid)
        
        # Bottom lip        
        if self.cb > 0.0:
            y_bot_lip = [-(0.5*hweb-(cb-tnom-r)),-(0.5*hweb)]                            
            ax.vlines([b_bot-1.5*tnom,b_bot-0.5*tnom],y_bot_lip[0],y_bot_lip[1],colors='k',linewidth=lw)
            ax.vlines([b_bot-tnom],y_bot_lip[0],y_bot_lip[1],colors='b',linewidth=lw_c,linestyle='-.')
            
            # Corner of the lip
            x_bot_left = (b_bot-1.5*tnom-r,-(0.5*h-tnom-r))
            
            bot_left_corner_in = patches.Arc(x_bot_left,2*r,2*r,0,270,360,color='k',linewidth=lw)
            bot_left_corner_out = patches.Arc(x_bot_left,2*(r+tnom),2*(r+tnom),0,270,360,color='k',linewidth=lw)
            bot_left_corner_mid = patches.Arc(x_bot_left,2*r+tnom,2*r+tnom,0,270,360,color='b',linewidth=lw_c,linestyle='-.')
            
            ax.add_patch(bot_left_corner_in)
            ax.add_patch(bot_left_corner_out)
            ax.add_patch(bot_left_corner_mid)
            
            ax.hlines(-(0.5*hweb-(cb-tnom-r)),b_bot-0.5*tnom,b_bot-1.5*tnom,colors='k',linewidth=lw)
                
        ax.set_aspect('equal')
    
    def export_cufsm(self,n=4):
        """ Exports the cross-section to be analyzed with CUFSM5 software
            run in Matlab            
        """
        print("Material:")
        print(f"100 {self.material.E:7.2f} {self.material.E:7.2f} {self.material.nu:3.1f} {self.material.nu:3.1f} {self.material.G:7.2f}")
        
        cufsm_nodes = []
        
        # Number of points; used as an index
        npt = 0
                
        h = self.h
        hweb = self.hw
        
        tnom = self.t_nom
        rm = self.Rm
        ro = self.Ro
        r = self.r
        
        bp_web = self.hw_center
        bp_bottom = self.bb_center
        bp_top = self.bt_center
        bp_bottom_lip = self.bb_lip_center
        bp_top_lip = self.bt_lip_center
        web_nodes = [[0.0,-0.5*bp_web],[0.0,0.5*bp_web]]
        
        if bp_bottom_lip > 0.0:
            Y = np.linspace(web_nodes[0][1]+bp_bottom_lip,web_nodes[0][1]+bp_bottom_lip-self.cb_straight,n+1)
            bottom_lip_nodes = [[bp_bottom,y] for y in Y]
                        
            #bottom_lip_nodes = [[bp_bottom,web_nodes[0][1]+bp_bottom_lip],
            #                    [bp_bottom,web_nodes[0][1]+bp_bottom_lip-0.5*self.cb_straight]]
            print(bottom_lip_nodes)
            for node in bottom_lip_nodes:
                cufsm_nodes.append(node)
            
            # Corner points
            # Lip-to-flange corner, bottom
            x_bottom_lip = (bp_bottom-rm,web_nodes[0][1]+rm)
        
            newPts = corner_points(x_bottom_lip,rm,phi=[0,-0.5*np.pi],n=n+1)
        
            for newPt in newPts[1:-1]:            
                cufsm_nodes.append(newPt)
            
            # The second and third nodes are for helping to cope
            # with the effective cross section.
            X = np.linspace(bp_bottom-rm,bp_bottom-rm-self.c_bottom,n+1)            
        else:
            X = np.linspace(bp_bottom,bp_bottom-self.c_bottom,n+1)            
        
        bottom_nodes = [[x,web_nodes[0][1]] for x in X]       
        
        for node in bottom_nodes:
            cufsm_nodes.append(node)
                        
        # Corner points
        # flange-to-web corner, bottom
        x_bottom_web = (rm,web_nodes[0][1]+rm)
        
        newPts = corner_points(x_bottom_web,rm,phi=[1.5*np.pi,np.pi],n=n+1)
        
        for newPt in newPts[1:-1]:            
            cufsm_nodes.append(newPt)
                
                
        # Web points
        Y = np.linspace(-0.5*bp_web+rm,0.5*bp_web-rm,n+1)
        web = [[0,y] for y in Y]
        
        for node in web:
            cufsm_nodes.append(node)
                
        
        # Corner points
        # web-to-flange corner, top
        x_top_web = (+rm,web_nodes[1][1]-rm)
        
        newPts = corner_points(x_top_web,rm,phi=[np.pi,0.5*np.pi],n=n+1)
        
        for newPt in newPts[1:-1]:            
            cufsm_nodes.append(newPt)
                        
        
        # Top flange        
        #top_nodes = [[rm,web_nodes[1][1]]]
        
        if bp_top_lip > 0.0:
            X = np.linspace(rm,bp_top-rm,n+1)
            
            top_nodes = [[x,web_nodes[1][1]] for x in X]            
            
            # Corner
            # web-to-flange corner, top
            x_top_lip = (bp_top-rm,web_nodes[1][1]-rm)
        
            for node in top_nodes:
                cufsm_nodes.append(node)
                            
        
            
            newPts = corner_points(x_top_lip,rm,phi=[0.5*np.pi,0],n=n+1)
        
            for newPt in newPts[1:-1]:            
                cufsm_nodes.append(newPt)
            
            # Top lip nodes
            Y = np.linspace(web_nodes[1][1]-rm,web_nodes[1][1]-bp_top_lip,n+1)
            top_lip_nodes = [[bp_top,y] for y in Y]
            
            for node in top_lip_nodes:
                cufsm_nodes.append(node)
            
            
        else:
            X = np.linspace(rm,bp_top,n+1)
            top_nodes = [[x,web_nodes[1][1]] for x in X]
        
            for node in top_nodes:
                cufsm_nodes.append(node)
        
        for p in cufsm_nodes:
            p[1] += 0.5*bp_web
        
        cuprof = OpenProf(cufsm_nodes,self.t)   
        cuprof.export_cufsm()
        cuprof.draw()
        
    
def Cexample(t=2.0,ca=0,cb=0):
    
    c = CSection(t_nom=t,h=150.0,a=65,b=61,ca=ca,cb=cb,r=3.0,t_coat=0.04)
    #c = CSection(t_nom=t,h=150.0,a=47,b=41,ca=ca,cb=cb,r=3.0,t_coat=0.04)
    #c = CSection(t_nom=t,h=64.0,a=34,b=34,ca=8,cb=8,r=1.0,t_coat=0.04)
    
    #c.Med = 10.0
    c.Ned = -10.0
    #c.effective_width_flange(psi=1.0, flange='top',verb=True)
    c.effective_section('compression',True)
    #c.effective_section('bending_y_pos',True)
    
    c.export_cufsm(n=4)
    #c.create_model()
    #c.prof.draw(node_labels=True)
    #c.create_effective_model()
    
    #c.prof_eff.draw(seg_labels=True,node_labels=True)

    #ygc, zgc = c.centroid_eff()
    #print(ygc,zgc)

    return c

def Cdesign(t=2.0,ca=0,cb=0):
    
    c = CSection(t_nom=t,h=150.0,a=47,b=47,ca=ca,cb=cb,r=3.0,t_coat=0.04)
    
    c.draw()
    
    #c.Med = 10.0
    #c.Ned = -10.0
    #c.effective_width_flange(psi=1.0, flange='top',verb=True)
    #c.effective_section('compression',True)
    """
    print("Axial resistance: {0:4.2f} kN".format(c.NRd*1e-3))
    
    c.prof.draw(coord_axes=True)
    plt.grid(True)
    c.prof_eff.draw(coord_axes=True)
    plt.grid(True)
    """
    return c

if __name__ == "__main__":
    
    #c = Cexample(t=0.7,ca=16,cb=16)
    #c = Cexample(t=1.5,ca=18,cb=18)
    #c = Cdesign(t=1.5,ca=18,cb=18)
    
    #z = ZSection(t_nom=1.5,h=200,a=74,b=66,ca=21.2,cb=21.2,r=3,
    #             material="S350GD",t_coat=0.04)
    
    #z = ruukki_z(300,2.5)
    
    #z.prof.draw(node_labels=True)
    
    #print(z.Agross()*1e-2)
    
    #c = CSection(1.0,150,47.0,41.0,16,16,3)
    #c.prof.draw(node_labels=False)
    
    c = CSection(1.5,150,47,41,18,18,3)
    c.Med = 30e6
    #c.Ned = -30.11e3
    g = c.centroid()
    print(g)
    c.prof.draw(node_labels=True)
    
    c.effective_section(load='bending_y_pos',verb=True)
    geff = c.centroid_eff()
    print(geff)
    c.prof_eff.draw(node_labels=True)
    
    """
    z = ZSection(ca=0.0,cb=0.0)
    c = CSection(1.48,97.3,37.0,37.0,12.5,12.5,0.85,t_coat=0.0)
    c.material.fy = 505.0
    #c.draw()
    z.create_model()
    z.prof.draw()
    c.create_model()
    #c.prof.draw()
    c.effective_width_lip(1.0,'top',verb=True)
    c.effective_width_lip(1.0,'bottom',verb=True)
    c.effective_width_flange(1.0,'top',verb=True)
    c.effective_width_flange(1.0,'bottom',verb=True)
    c.effective_width_web(1.0,verb=True)
    """


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 14 19:46:57 2019

@author: kmela
"""

from sections.steel.catalogue import ipe_profiles
from sections.steel.ISection import IPE

for prof_name, dims in ipe_profiles.items():
    p = IPE(int(dims["h"]))
    
    